# -*- coding: utf-8 -*-

"""scinstr.tctrl
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2018-2025
license   GPL v3.0+
brief     Base class to handle Lakeshore 350 device.
"""

import logging
import socket
import time
import serial

# Ethernet specific
PORT = 7777
ETH_TIMEOUT = 1.0  # in second

# USB specific
BAUDRATE = 57600
DATA_BITS = serial.SEVENBITS
STOP_BIT = serial.STOPBITS_ONE
ODD_PARITY = serial.PARITY_ODD
USB_TIMEOUT = 1.0  # 0 is non-blocking mode

CMD = ('CLS', '*ESE', '*ESE?', '*ESR?', '*IDN?', '*OPC', '*OPC?', '*RST',
       '*SRE', '*SRE?', '*STB?', '*TST?', '*WAI', 'ALARM', 'ALARM?',
       'ALARMST?', 'ALMRST', 'ANALOG', 'ANALOG?', 'AOUT?', 'ATUNE', 'BRIGT',
       'BRIGT?', 'CRDG?', 'CRVDEL', 'CRVHDR', 'CRVHDR?', 'CRVPT', 'CRVPT?',
       'DFLT', 'DIOCUR', 'DISPFLD', 'DISPFLD?', 'DISPLAY', 'DISPLAY?',
       'FILTER', 'FILTER?', 'HTR?', 'HTRSET', 'HTRSET?', 'HTRST?', 'IEEE',
       'IEEE?', 'INCRV', 'INCRV?', 'INNAME', 'INNAME?', 'INTSEL', 'INTSEL?',
       'INTYPE', 'INTYPE?', 'KRDG?', 'LEDS', 'LEDS?', 'LOCK', 'LOCK?',
       'MDAT?', 'MNMXRST', 'MODE', 'MODE?', 'MOUT', 'MOUT?', 'NET', 'NET?',
       'NETID?', 'OPST?', 'OPSTE', 'OPSTE?', 'OPSTR?', 'OUTMODE', 'OUTMODE?',
       'PID', 'PID?', 'RAMP', 'RAMP?', 'RAMPST?', 'RANGE', 'RANGE?', 'RDGST?',
       'RELAY', 'RELAY?', 'RELAYST?', 'SCAL', 'SETP', 'SETP?', 'SRDG?',
       'TEMP?', 'TLIMIT', 'TLIMIT?', 'TUNEST?', 'WARMUP', 'WARMUP?', 'WEBLOG',
       'WEBLOG?', 'ZONE', 'ZONE?')


# =============================================================================
class L350():
    """Base class to handle Lakeshore 350 device.
    """

    def connect(self):
        """Abstract protocol connect process. Derived classes must implement
        the connect process dedicated to the specific protocol used.
        :returns: None
        """
        raise NotImplementedError("Method not implemented by derived class")

    def close(self):
        """Abstract protocol closing process. Derived classes must implement
        the closing process dedicated to the specific protocol used.
        :returns: None
        """
        raise NotImplementedError("Method not implemented by derived class")

    def _write(self, data):
        """Abstract protocol write process. Derived classes must implement
        the write process dedicated to the specific protocol used.
        :param data: data writes to device (str)
        :returns: None
        """
        raise NotImplementedError("Method not implemented by derived class")

    def _read(self, length):
        """Abstract protocol read process. Derived classes must implement
        the read process dedicated to the specific protocol used.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        raise NotImplementedError("Method not implemented by derived class")

    def read(self, length=None):
        """A basic read method: read a message from device.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        try:
            retval = self._read(length)
        except Exception as ex:
            logging.error("Read error: %r", ex)
            return ''
        logging.debug("read: %r", retval)
        return retval

    def write(self, data):
        """A basic write method: writes "data" to device.
        :param data: data writes to device (str)
        :returns: None
        """
        try:
            nb_bytes = self._write(data)
        except Exception as ex:
            logging.error("Write error: %r", ex)
            return 0
        logging.debug("write: %r", data)
        return nb_bytes

    def query(self, data, length=1024):
        """Write than read procedure.
        :param data: data writes to device (str)
        :returns: Message returned by device (str)
        """
        self.write(data)
        return self.read(length)

    def local(self):
        """Set device in local mode.
        :returns: None
        """
        self.write("MODE 0")

    def remote(self, lock=False):
        """Set device in remote mode. If lock is True, device is in remote
        mode with local lockout.
        :param lock: if True, remote with local lockout (bool)
        :returns: None
        """
        if lock is True:
            self.write("MODE 2")
        else:
            self.write("MODE 1")

    @property
    def idn(self):
        """Return product id of device.
        :returns: product id of device (str)
        """
        _id = self.query("*IDN?")
        return _id

    def get_rc(self, key: int) -> float:
        """Return resistance value with respect to key:
        - 0 -> Channel A
        - 1 -> Channel B
        - 2 -> Channel C
        - 3 -> Channel D
        - 4 -> Heater 1
        - 5 -> Heater 2
        """
        if key == 0:
            return float(self.query('SRDG? A', 128))
        elif key == 1:
            return float(self.query('SRDG? B', 128))
        elif key == 2:
            return float(self.query('SRDG? C', 128))
        elif key == 3:
            return float(self.query('SRDG? D', 128))
        elif key == 4:
            return float(self.query('HTR? 1', 128))
        elif key == 5:
            return float(self.query('HTR? 2', 128))
        else:
            raise ValueError("Key value must be beetween 0 and 5.")

    def get_tc(self, key: int) -> float:
        """Return temperature value with respect to key:
        - 0 -> Channel A
        - 1 -> Channel B
        - 2 -> Channel C
        - 3 -> Channel D
        - 4 -> Heater
        """
        if key == 0:
            return float(self.query('KRDG? A', 128))
        elif key == 1:
            return float(self.query('KRDG? B', 128))
        elif key == 2:
            return float(self.query('KRDG? C', 128))
        elif key == 3:
            return float(self.query('KRDG? D', 128))
        elif key == 4:
            return float(self.query('HTR? 1', 128))
        elif key == 5:
            return float(self.query('HTR? 2', 128))
        else:
            raise ValueError("Key value must be beetween 0 and 5.")

    def get_rc_all(self) -> list[float]:
        """Return all resistances and all heater values.
        """
        retval = self.query("SRDG? A; SRDG? B;SRDG? C;SRDG? D;HTR? 1;HTR? 2", 512)
        return [float(value) for value in retval.split(';')]

    def get_tc_all(self) -> list[float]:
        """Return all temperatures and all heater values.
        """
        retval = self.query("KRDG? A; KRDG? B;KRDG? C;KRDG? D;HTR? 1;HTR? 2", 512)
        return [float(value) for value in retval.split(';')]


# =============================================================================
class L350Eth(L350):
    """Handle Lakeshore 350 device through ethernet interface.
    """

    def __init__(self, ip='', port=PORT, timeout=ETH_TIMEOUT):
        """'Constructor'.
        :param ip: IP address of device (str)
        :param port: Device port in use (int)
        :param timeout: Timeout in second (float)
        :returns: None
        """
        self._sock = None
        self.ip = ip
        self.port = port
        self.timeout = timeout

    def connect(self):
        """Open connection with device.
        :returns: True if connection succeed False elsewhere (bool)
        """
        self._sock = socket.socket(socket.AF_INET,
                                   socket.SOCK_STREAM,
                                   socket.IPPROTO_TCP)
        self._sock.settimeout(self.timeout)
        try:
            self._sock.connect((self.ip, self.port))
        except ValueError as ex:
            logging.error("Wrong connection parameters: %r", ex)
            return False
        except socket.timeout:
            logging.error("Timeout on connection")
            return False
        logging.info("Connected to Lakeshore")
        return True

    def close(self):
        """Close connection with device.
        :returns: None
        """
        if self._sock is None:
            return
        self._sock.close()
        self._sock = None
        logging.info("Connection to Lakeshore closed")

    def _write(self, data):
        """Specific ethernet writing process.
        :param data: data writes to device (str)
        :returns: number of bytes sent (int)
        """
        time.sleep(0.01)  # Device seems temporisation between com
        return self._sock.send((data + '\n').encode('utf-8'))

    def _read(self, length=100):
        """Specific ethernet reading process.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        time.sleep(0.01)  # Device seems temporisation between com
        return self._sock.recv(length).decode('utf-8').strip('\r\n')


# =============================================================================
class L350Usb(L350):
    """Handle Lakeshore 350 device through USB interface.
    The USB interface emulates an RS-232 serial port with the folowing
    configuration parameters (see manual p.116):
    - Baud rate 57,600
    - Data bits 7
    - Start bits 1
    - Stop bits 1
    - Parity Odd
    - Flow control None
    - Handshaking None
    """

    def __init__(self, port='', timeout=USB_TIMEOUT):
        self._ser = None
        self.port = port
        self._timeout = timeout

    def connect(self):
        """Connect to the remote host.
        :returns: True if connection succeeded, False otherwise (bool)
        """
        self._ser = serial.Serial(self.port,
                                  baudrate=BAUDRATE,
                                  bytesize=DATA_BITS,
                                  parity=ODD_PARITY,
                                  stopbits=STOP_BIT,
                                  timeout=self._timeout)
        if self._ser.isOpen() is False:
            try:
                self._ser.open()
            except ValueError as ex:
                logging.error("Wrong connection parameters: %r", ex)
                return False
        logging.info("Connected to Lakeshore")
        return True

    def close(self):
        """Closes the underlying serial connection
        """
        if self._ser is not None:
            try:
                self._ser.close()
            except Exception as ex:
                logging.error("Error when closing USB connection: %r", ex)
        self._ser = None
        logging.info("Connection to Lakeshore closed")

    def _write(self, data):
        """Specific USB writing process.
        :param data: data writes to device (str)
        :returns: number of bytes sent (int)
        """
        time.sleep(0.05)  # 50 ms minimum between USB com (p.120-121 manual)
        return self._ser.write((data + '\n').encode('utf-8'))

    def _read(self, length=None):
        """Specific USB reading process.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        time.sleep(0.05)  # 50 ms minimum between USB com (p.120-121 manual)
        return self._ser.read_until(size=length).decode('utf-8').strip('\r\n')

    def hard_flush(self):
        """Try to make a "hard" flush of output data.
        Needed because commands can cause inappropriate behavior (CRVDEL)
        :returns: None
        """
        self._ser.flush()
        self._ser.reset_output_buffer()
        time.sleep(1.0)

    @property
    def timeout(self):
        """Gets timeout on socket operations.
        :returns: timeout value in second (float)
        """
        if self._ser is None:
            return self._timeout
        return self._ser.timeout

    @timeout.setter
    def timeout(self, timeout):
        """Sets timeout on socket operations.
        :param timeout: timeout value in second (float)
        :returns: None
        """
        self._timeout = timeout
        if self._ser is None:
            return
        self._ser.timeout = timeout



# -*- coding: utf-8 -*-

"""scinstr.vaccum
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2018-2024
license   GPL v3.0+
brief     Class to Pfeiffer TPG261 controller device.
"""

import serial
import scinstr.vacuum.gauge as gauge

ERR_CODES = {
    0: 'No error',
    1: 'Watchdog has responded',
    2: 'Task fail error',
    3: 'EPROM error',
    4: 'RAM error',
    5: 'EEPROM error',
    7: 'Display error',
    8: 'A/D converter error',
    9: 'Gauge 1 error (e.g. filament rupture, no supply)',
    10: 'Gauge 1 identification error',
    11: 'Gauge 2 error (e.g. filament rupture, no supply)',
    12: 'Gauge 2 identification error'
}

PRESSURE_READING_STATUS = {
    0: 'Measurement data okay',
    1: 'Underrange',
    2: 'Overrange',
    3: 'Sensor error',
    4: 'Sensor off',
    5: 'No sensor',
    6: 'Identification error'
}

GAUGE_TYPE = {
    'TPR': 'Pirani Gauge (all)',
    'IKR9': 'Cold Cathode Gauge 10-9',
    'IKR11': 'Cold Cathode Gauge 10-11',
    'PKR': 'FullRange CC Gauge',
    'PBR': 'FullRange BA Gauge',
    'IMR': 'Pirani / High Pressure Gauge',
    'CMR': 'Linear gauge',
    'noSEn': 'no SEnsor',
    'noid': 'no identifier'
}

FSR = {
    0: '0.01 mbar',
    1: '0.1 mbar',
    2: '1 mbar',
    3: '10 mbar',
    4: '100 mbar',
    5: '1000 mbar default',
    6: '2 bar',
    7: '5 bar',
    8: '10 bar',
    9: '50 bar'
}

SER_TIMEOUT = 0.2
ETH_TIMEOUT = 0.2


# =============================================================================
class Tpg261Serial(gauge.GaugeSerial):
    """Handle TPG261 vacuum gauge device through USB connection.
    The USB interface emulates an RS-232 serial port with the folowing
    configuration parameters:
    - Baud rate 9600
    - Start bits 1
    - Data bits 8
    - Parity None
    - Stop bits 1
    - Flow control None
    - Handshaking None
    """

    def __init__(self, port='', timeout=SER_TIMEOUT):
        super().__init__(port=port,
                         baudrate=9600,
                         bytesize=8,
                         parity=serial.PARITY_NONE,
                         stopbits=1,
                         timeout=timeout)


# =============================================================================
class Tpg261Eth(gauge.GaugeEth):
    """Handle TPG261 vacuum gauge device through ethernet interface.
    """

    PORT = 7777

    def __init__(self, ip='', port=PORT, timeout=ETH_TIMEOUT):
        super().__init__(ip, port, timeout)


# =============================================================================
if __name__ == '__main__':
    import logging
    import argparse

    DATE_FMT = "%d/%m/%Y %H:%M:%S"
    LOG_FORMAT = "%(asctime)s %(levelname) -8s %(filename)s " + \
                 " %(funcName)s (%(lineno)d): %(message)s"
    logging.basicConfig(level=logging.DEBUG,
                        datefmt=DATE_FMT,
                        format=LOG_FORMAT)

    PARSER = argparse.ArgumentParser(description='Test Pfeiffer tpg261')
    PARSER.add_argument('-p', '--port', dest='port', type=str,
                        help='Serial port to connect to')

    ARGS = PARSER.parse_args()
    PORT = ARGS.port
    INST = Tpg261Serial(port=PORT)
    INST.connect()

    RETVAL = INST.reset()
    if RETVAL != '0':
        INST.close()
        exit()

    print("GID: %s" % (INST.gauge_id))
    print("Error: %s" % (INST.error))

    INST.close()



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""package fswp
author    Benoit Dubois
copyright FEMTO ENGINEERING
license   GPL v3.0+
brief     Acquire data trace from FSWP device.
"""

import sys
import os.path as path
import logging
import datetime
import socket
import struct
import signal
from pyqtgraph.parametertree import Parameter, ParameterTree
import pyqtgraph as pg
import numpy as np
from PyQt5.QtCore import pyqtSlot, pyqtSignal, QDir, QFileInfo
from PyQt5.QtGui import QApplication, QMainWindow, QWidget, QVBoxLayout, \
    QHBoxLayout, QMessageBox, QFileDialog, QSplitter

# Ctrl-c closes the application
signal.signal(signal.SIGINT, signal.SIG_DFL)

CONSOLE_LOG_LEVEL = logging.DEBUG
FILE_LOG_LEVEL = logging.WARNING

IP = "172.16.120.200"
#PORT = 4880  # HiSLIP protocol port
PORT = 5025  # Raw socket port
APP_NAME = "FswpAcquire"


#===============================================================================
class FswpDev():

    def __init__(self, ip, port=PORT, timeout=1.0):
        """Constructor.
        :param ip: IP address of device (str)
        :param port: Ethernet port of device (int)
        :param timeout: Timeout on connection instance (float)
        :returns: None
        """
        self.ip = ip
        self.port = port
        self._timeout = timeout
        self._sock = None

    def connect(self):
        """Connect to device.
        """
        try:
            self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._sock.settimeout(self._timeout)
            self._sock.connect((self.ip, self.port))
        except ValueError as ex:
            logging.error("Connection parameters out of range: %r", ex)
            return False
        except socket.timeout:
            logging.error("Timeout on connection")
            return False
        except Exception as ex:
            logging.error("Unexpected exception during connection with " + \
                          "device: %r", ex)
            return False
        else:
            logging.debug("Connected to device")
            return True

    def write(self, data):
        """"Ethernet writing process.
        :param data: data writes to device (str)
        :returns: None
        """
        try:
            self._sock.send((data + '\n').encode('utf-8'))
        except socket.timeout:
            logging.error("Timeout")
        except Exception as ex:
            logging.error(ex)
        logging.debug("write " + data.strip('\n'))

    def read(self, length=100):
        """Read process.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        try:
            retval = self._sock.recv(length).decode('utf-8').strip('\n')
        except socket.timeout:
            logging.error("Timeout")
            return ''
        except Exception as ex:
            logging.error(ex)
            raise ex
        logging.debug("read: " + retval)
        return retval

    def raw_read(self, length=512):
        """Raw read process.
        :param length: length of message to read (int)
        :returns: Message reads from device (bytes)
        """
        try:
            data = self._sock.recv(length)
        except socket.timeout:
            logging.error("Timeout")
            return bytes()
        except Exception as ex:
            logging.error(ex)
            raise ex
        return data

    def bin_read(self):
        """Read binary data then decode them to ascii.
        The reading process is specific to the transfert of binary data
        with these VNA devices: <header><data><EOT>, with:
        - <header>: #|lenght of bytes_count (one byte)|bytes_count
        - <data>: "REAL,32" (float32) binary data
        - <EOT>: '\n' character.
        Note: The data transfert format must be selected to "REAL,32"" before
        using this method
        """
        header_max_length = 11
        raw_data = self.raw_read(header_max_length)
        if raw_data.find(b'#') != 0:
            logging.error("Data header not valid")
            return
        byte_count_size = int(raw_data[1:2])
        byte_count = int(raw_data[2:2+byte_count_size])
        # Note : Read 'byte_count' bytes but only
        # 2 + byte_count_nb + byte_count - header_max_length
        # needs to be readen.
        # This tip can be used because EOF ('\n') is transmited at the end of
        # the message and thus stop reception of data.
        while len(raw_data) < byte_count:
            raw_data += self.raw_read(byte_count)
        nb_elem = int(byte_count / 4)
        data = np.asarray(struct.unpack("<{:d}f".format(nb_elem),
                                        raw_data[2+byte_count_size:-1]))
        return data

    def query(self, msg, length=100):
        """Basic query process: write query then read response.
        """
        self.write(msg)
        return self.read(length)

    def reset(self):
        """Reset device.
        """
        self.write("*RST")

    @property
    def id(self):
        """Return ID of device.
        """
        return self.query("*IDN?")

    def get_data_trace(self, trace=1, window=1):
        """Get a data trace.
        :param :
        :return: Array of measurement data.
        """
        self.write("FORM REAL, 32")
        self.write("TRAC{}? TRACE{}".format(window, trace))
        data = self.bin_read()
        datax = data[::2]
        datay = data[1::2]
        return np.asarray([datax, datay])

    def get_crossco_gain_indicator(self, window=1):
        """Get the data of the cross-correlation gain indicator (grey area).
        :param :
        :return: Array of the cross-correlation gain indicator data.
        """
        self.write("FORM REAL, 32")
        self.write("TRAC{}? XGINdicator".format(window))
        data = self.bin_read()
        datax = data[::2]
        datay = data[1::2]
        return np.asarray([datax, datay])
    


#===============================================================================
MAIN_PARAMS = [
    {'name': 'Load data', 'type': 'group', 'children': [
        {'name': 'From device', 'type': 'group', 'children': [
            {'name': 'IP', 'type': 'str', 'value': IP},
            {'name': 'Port', 'type': 'int', 'value': PORT},
            {'name': 'Trace', 'type': 'group', 'children': [
                {'name': 'Number', 'type': 'int', 'value': 1},
                {'name': 'Bench noise', 'type': 'bool', 'value': False},
            ]},
            
            {'name': 'Acquisition', 'type': 'action'},
        ]},
        {'name': 'From file', 'type': 'group', 'children': [
            {'name': 'Filename', 'type': 'str'},
            {'name': 'Open', 'type': 'action'},
        ]},
    ]},
    {'name': 'Plot', 'type': 'group', 'children': []},
]

CURVE_PARAMS = [
                {'name': 'Enable', 'type': 'bool', 'value': True},
                {'name': 'Scale', 'type': 'float', 'value': 1.0},
                {'name': 'Offset', 'type': 'float', 'value': 0.0},
                {'name': 'Color', 'type': 'color'},
        ]


class FswpUi(QMainWindow):
    """Ui of extract peak application.
    """

    def __init__(self):
        """Constructor.
        :returns: None
        """
        super().__init__()
        self.setWindowTitle("FSWP")
        self.setCentralWidget(self._central_widget())
        self.plot.showGrid(True, True, 0.5)
        self.plot.setLogMode(x=True, y=False)

    def _central_widget(self):
        """Generates central widget.
        :returns: central widget of UI (QWidget)
        """
        self.p = Parameter.create(name='params',
                                  type='group',
                                  children=MAIN_PARAMS)
        self.ptree = ParameterTree()
        self.ptree.setParameters(self.p, showTop=False)
        self.ptree.resizeColumnToContents(1)
        self.ptree.resizeColumnToContents(2)
        self.plot = pg.PlotWidget()
        splitter = QSplitter(self)
        splitter.addWidget(self.ptree)
        splitter.addWidget(self.plot)
        main_layout = QHBoxLayout()
        main_layout.addWidget(splitter)
        central_widget = QWidget()
        central_widget.setLayout(main_layout)
        return central_widget


#===============================================================================
class FswpApp(QApplication):
    """Fswp application.
    """

    # Emit data key value
    get_from_dev_done = pyqtSignal(str)
    get_from_file_done = pyqtSignal(str)

    def __init__(self, args):
        """Constructor.
        :returns: None
        """
        super().__init__(args)
        self._ui = FswpUi()
        self._data = {}
        self._ui.p.param(
            'Load data', 'From device', 'Acquisition').sigActivated.connect(
                self.get_data_from_device)
        self._ui.p.param(
            'Load data', 'From file', 'Open').sigActivated.connect(
                self.get_data_from_file)
        self.get_from_dev_done.connect(self.save_data)
        self.get_from_dev_done.connect(self.handle_acq_data)
        self.get_from_dev_done.connect(self.display_data)
        self.get_from_file_done.connect(self.handle_acq_data)
        self.get_from_file_done.connect(self.display_data)
        self._ui.p.param('Plot').sigTreeStateChanged.connect(
            self.display_data)
        self._ui.show()

    @pyqtSlot()
    def get_data_from_file(self):
        filename = QFileDialog(self._ui).getOpenFileName(
            parent=None,
            caption="Choose file to load",
            directory=QDir.currentPath(),
            filter="Any files (*)")[0]
        if filename == '':
            return
        try:
            data = np.loadtxt(filename)
        except Exception as ex:
            logging.error("Problem when reading file: %s", str(ex))
            QMessageBox.warning(self._ui, "Acquisition problem",
                                "Problem when reading file: {}".format(ex),
                                QMessageBox.Ok)
            return
        key = QFileInfo(filename).baseName()
        self._data[key] = data
        self.get_from_file_done.emit(key)

    @pyqtSlot()
    def get_data_from_device(self):
        """Acquire data process.
        :returns: None
        """
        ip = self._ui.p.param('Load data', 'From device', 'IP').value()
        port = self._ui.p.param('Load data', 'From device', 'Port').value()
        trace_nb = self._ui.p.param('Load data',
                                    'From device',
                                    'Trace',
                                    'Number').value()
        bench_noise = self._ui.p.param('Load data',
                                       'From device',
                                       'Trace',
                                       'Bench noise').value()
        # Acquisition itself
        dev = FswpDev(ip, port)
        if not dev.connect():
            logging.error("Connection error")
            return
        try:
            data = dev.get_data_trace(trace=trace_nb)
            if bench_noise is True:
                bnoise = [dev.get_crossco_gain_indicator()[0]]
                data = np.concatenate((data, bnoise))
        except Exception as ex:
            logging.error("Problem during acquisition: %r", ex)
            QMessageBox.warning(self._ui, "Acquisition problem",
                                "Problem during acquisition: {}".format(ex),
                                QMessageBox.Ok)
            return
        now = datetime.datetime.utcnow()
        key = now.strftime("%Y%m%d-%H%M%S")
        self._data[key] = data
        self.get_from_dev_done.emit(key)

    @pyqtSlot(str)
    def handle_acq_data(self, key):
        self._ui.p.param('Plot').addChild(
            Parameter.create(name=key,
                             type='group',
                             children=CURVE_PARAMS))
        # Get a new color for the new curve
        ##for child in self._ui.p.param('Plot').children():
        ##    color =  child.param('Color').value()

    @pyqtSlot(str)
    def save_data(self, key):
        np.savetxt("{}.dat".format(key), self._data[key])

    @pyqtSlot()
    def display_data(self):
        for pdi in self._ui.plot.getPlotItem().listDataItems():
            self._ui.plot.getPlotItem().removeItem(pdi)
        for child in self._ui.p.param('Plot').children():
            if child.param('Enable').value() is True:
                scale = child.param('Scale').value()
                offset = child.param('Offset').value()
                color =  child.param('Color').value()
                self._ui.plot.plot(
                    self._data[child.name()][0],
                    self._data[child.name()][1] * scale + offset,
                    pen=color)


#==============================================================================
def configure_logging():
    """Configures logs.
    """
    home = path.expanduser("~")
    log_file = "." + APP_NAME + ".log"
    abs_log_file = path.join(home, log_file)
    date_fmt = "%d/%m/%Y %H:%M:%S"
    log_format = "%(asctime)s %(levelname) -8s %(filename)s " + \
                 " %(funcName)s (%(lineno)d): %(message)s"
    logging.basicConfig(level=FILE_LOG_LEVEL, \
                        datefmt=date_fmt, \
                        format=log_format, \
                        filename=abs_log_file, \
                        filemode='w')
    console = logging.StreamHandler()
    # define a Handler which writes messages to the sys.stderr
    console.setLevel(CONSOLE_LOG_LEVEL)
    # set a format which is simpler for console use
    console_format = '%(levelname) -8s %(filename)s (%(lineno)d): %(message)s'
    formatter = logging.Formatter(console_format)
    # tell the handler to use this format
    console.setFormatter(formatter)
    # add the handler to the root logger
    logging.getLogger('').addHandler(console)

#==============================================================================
def main():
    configure_logging()
    app = FswpApp(sys.argv)
    sys.exit(app.exec_())

#==============================================================================
if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""HTTP server for modbus simulator.

The modbus simulator contain 3 distinct parts:

- Datastore simulator, to define registers and their behaviour including actions: (simulator)(../../datastore/simulator.py)
- Modbus server: (server)(./http_server.py)
- HTTP server with REST API and web pages providing an online console in your browser

Multiple setups for different server types and/or devices are prepared in a (json file)(./setup.json), the detailed configuration is explained in (doc)(README.rst)

The command line parameters are kept to a minimum:

usage: main.py [-h] [--modbus_server MODBUS_SERVER]
               [--modbus_device MODBUS_DEVICE] [--http_host HTTP_HOST]
               [--http_port HTTP_PORT]
               [--log {critical,error,warning,info,debug}]
               [--json_file JSON_FILE]
               [--custom_actions_module CUSTOM_ACTIONS_MODULE]

Modbus server with REST-API and web server

options:
  -h, --help            show this help message and exit
  --modbus_server MODBUS_SERVER
                        use <modbus_server> from server_list in json file
  --modbus_device MODBUS_DEVICE
                        use <modbus_device> from device_list in json file
  --http_host HTTP_HOST
                        use <http_host> as host to bind http listen
  --http_port HTTP_PORT
                        use <http_port> as port to bind http listen
  --log {critical,error,warning,info,debug}
                        set log level, default is info
  --log_file LOG_FILE
                        name of server log file, default is "server.log"
  --json_file JSON_FILE
                        name of json_file, default is "setup.json"
  --custom_actions_module CUSTOM_ACTIONS_MODULE
                        python file with custom actions, default is none
"""
import argparse
import asyncio
import os

from pymodbus import pymodbus_apply_logging_config
from pymodbus.logging import Log
from pymodbus.server.simulator.http_server import ModbusSimulatorServer


def get_commandline(extras=None, cmdline=None):
    """Get command line arguments."""
    parser = argparse.ArgumentParser(
        description="Modbus server with REST-API and web server"
    )
    parser.add_argument(
        "--modbus_server",
        help="use <modbus_server> from server_list in json file",
        type=str,
    )
    parser.add_argument(
        "--modbus_device",
        help="use <modbus_device> from device_list in json file",
        type=str,
    )
    parser.add_argument(
        "--http_host",
        help="use <http_host> as host to bind http listen",
        type=str,
    )
    parser.add_argument(
        "--http_port",
        help="use <http_port> as port to bind http listen",
        type=str,
        default=8081,
    )
    parser.add_argument(
        "--log",
        choices=["critical", "error", "warning", "info", "debug"],
        help="set log level, default is info",
        default="info",
        type=str,
    )
    parser.add_argument(
        "--json_file",
        help='name of json file, default is "setup.json"',
        type=str,
        default=os.path.join(os.path.dirname(__file__), "setup.json"),
    )
    parser.add_argument(
        "--log_file",
        help='name of server log file, default is "server.log"',
        type=str,
    )
    parser.add_argument(
        "--custom_actions_module",
        help="python file with custom actions, default is none",
        type=str,
    )
    if extras:
        for extra in extras:
            parser.add_argument(extra[0], **extra[1])
    args = parser.parse_args(cmdline)
    pymodbus_apply_logging_config(args.log.upper())
    Log.info("Start simulator")
    cmd_args = {}
    for argument in args.__dict__:
        if argument == "log":
            continue
        if args.__dict__[argument] is not None:
            cmd_args[argument] = args.__dict__[argument]
    return cmd_args


async def run_main():
    """Run server async."""
    cmd_args = get_commandline()
    task = ModbusSimulatorServer(**cmd_args)
    await task.run_forever()


def main():
    """Run server."""
    asyncio.run(run_main(), debug=True)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""package scinstr.lockin
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2018
license   GPL v3.0+
brief     Class for handling SR8x0 device throuch serial connection.
details   Use FTDI USB to serial RS232 adapter (USB-RS232-WE-1800-BT_0.0).
          Lockin exhibits sub-D25 female connector.
          Connection between cable and DE25:
            orange (TXD) <-> 2
            yellow (RXD) <-> 3
            black  (GND) <-> 7
          Rear view (solder face) of DB25 connector:
            13 12 11 10 9 8 Black 6 5 Yellow Orange 1
              25 24 23 22 21 20 19 18 17 16 15 14
"""

import logging
import serial
import scinstr.lockin.sr8x0 as sr8x0


class Sr8x0Serial(sr8x0.Sr8x0):
    """Stanford Research System SR8x0 serial interface.
    """

    DEFAULT_TIMEOUT = 1.0

    def __init__(self, port=None, baudrate=9600, parity=serial.PARITY_NONE,
                 rtscts=False, dsrdtr=False, timeout=DEFAULT_TIMEOUT):
        """The constructor.
        """
        super().__init__()
        self.port = port
        self._baudrate = baudrate
        self._parity = parity
        self._rtscts = rtscts
        self._dsrdtr = dsrdtr
        self._timeout = timeout
        self._ser = None

    def connect(self):
        """Connect to the remote host
        :returns: True if connection succeeded, False otherwise
        """
        self._ser = serial.Serial(port=self.port,
                                  baudrate=self._baudrate,
                                  parity=self._parity,
                                  rtscts=self._rtscts,
                                  dsrdtr=self._dsrdtr,
                                  timeout=self._timeout)
        if self._ser.isOpen() is False:
            try:
                self._sock.connect()
            except ValueError as ex:
                logging.error("Wrong connection parameters: %r", ex)
                return False
            except serial.timeout:
                logging.error("Timeout on connection")
                return False
        logging.info("Connected to SR8x0")
        return True

    def close(self):
        """Closes the underlying serial connection
        """
        if self._ser is not None:
            try:
                self._ser.close()
            except Exception as ex:
                logging.error("Error when closing USB connection: %r", ex)
        self._ser = None
        logging.info("Connection to SR8x0 closed")

    def write(self, data):
        """Specific serial writing process.
        :param data: data writes to device (str)
        :returns: number of bytes sent (int)
        """
        try:
            nb_bytes = self._ser.write((data + '\n').encode('utf-8'))
        except serial.SerialTimeoutException:
            logging.error("Device write timeout")
            return 0
        logging.debug("write %s", data)
        return nb_bytes

    def read(self, length):
        """Specific serial reading process.
        :param length: length of message to read (int)
        :returns: Message reads from device (str)
        """
        try:
            retval = self._ser.read(length).decode('utf-8').strip('\n')
        except serial.SerialTimeoutException:
            logging.error("Device read timeout")
            return ''
        logging.debug("read %s", retval)
        return retval


# =============================================================================
if __name__ == '__main__':
    import sys
    import argparse

    # Handles log
    DATE_FMT = "%d/%m/%Y %H:%M:%S"
    LOG_FORMAT = "%(asctime)s %(levelname) -8s %(filename)s " + \
                 " %(funcName)s (%(lineno)d): %(message)s"
    logging.basicConfig(level=logging.DEBUG,
                        datefmt=DATE_FMT,
                        format=LOG_FORMAT)

    PARSER = argparse.ArgumentParser(description='Lakeshore Monitoring')
    PARSER.add_argument('--port', dest='port', type=str,
                        help='If interface = \'usb\', serial port')
    ARGS = PARSER.parse_args()
    PORT = ARGS.port

    INST = Sr8x0Serial(port=PORT, baudrate=9600, timeout=1.0)
    RETVAL = INST.connect()
    if RETVAL is False:
        sys.exit("Connection error")

    print("IDN: %s" % INST.idn)

    print("Single shot acquisition")
    INST.reset_data_storage()
    INST.sample_rate = sr8x0.SAMPLE_RATES["256 Hz"]
    INST.single_shot = 1
    INST.start_data_storage()
    INST.wait_bit0()
    DATA = INST.read_buffer(_format='a')
    with open("sr810_data.dat", 'w') as fd:
        fd.write("# Sample rate: " + INST.sample_rate)
        for i in DATA:
            fd.write("%e\n" % i)
    INST.close()



"""Modbus client async TCP communication."""
from __future__ import annotations

import select
import socket
import time
from collections.abc import Callable
from ssl import SSLWantReadError

from pymodbus.client.base import ModbusBaseClient, ModbusBaseSyncClient
from pymodbus.exceptions import ConnectionException
from pymodbus.framer import FramerType
from pymodbus.logging import Log
from pymodbus.pdu import ModbusPDU
from pymodbus.transport import CommParams, CommType


class AsyncModbusTcpClient(ModbusBaseClient):
    """**AsyncModbusTcpClient**.

    Fixed parameters:

    :param host: Host IP address or host name

    Optional parameters:

    :param framer: Framer name, default FramerType.SOCKET
    :param port: Port used for communication
    :param name: Set communication name, used in logging
    :param source_address: source address of client
    :param reconnect_delay: Minimum delay in seconds.milliseconds before reconnecting.
    :param reconnect_delay_max: Maximum delay in seconds.milliseconds before reconnecting.
    :param timeout: Timeout for connecting and receiving data, in seconds.
    :param retries: Max number of retries per request.
    :param trace_packet: Called with bytestream received/to be sent
    :param trace_pdu: Called with PDU received/to be sent
    :param trace_connect: Called when connected/disconnected

    .. tip::
        The trace methods allow to modify the datastream/pdu !

    .. tip::
        **reconnect_delay** doubles automatically with each unsuccessful connect, from
        **reconnect_delay** to **reconnect_delay_max**.
        Set `reconnect_delay=0` to avoid automatic reconnection.

    Example::

        from pymodbus.client import AsyncModbusTcpClient

        async def run():
            client = AsyncModbusTcpClient("localhost")

            await client.connect()
            ...
            client.close()

    Please refer to :ref:`Pymodbus internals` for advanced usage.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        host: str,
        *,
        framer: FramerType = FramerType.SOCKET,
        port: int = 502,
        name: str = "comm",
        source_address: tuple[str, int] | None = None,
        reconnect_delay: float = 0.1,
        reconnect_delay_max: float = 300,
        timeout: float = 3,
        retries: int = 3,
        trace_packet: Callable[[bool, bytes], bytes] | None = None,
        trace_pdu: Callable[[bool, ModbusPDU], ModbusPDU] | None = None,
        trace_connect: Callable[[bool], None] | None = None,
    ) -> None:
        """Initialize Asyncio Modbus TCP Client."""
        if not hasattr(self,"comm_params"):
            if framer not in [FramerType.SOCKET, FramerType.RTU, FramerType.ASCII]:
                raise TypeError("Only FramerType SOCKET/RTU/ASCII allowed.")
            self.comm_params = CommParams(
                comm_type=CommType.TCP,
                host=host,
                port=port,
                comm_name=name,
                source_address=source_address,
                reconnect_delay=reconnect_delay,
                reconnect_delay_max=reconnect_delay_max,
                timeout_connect=timeout,
            )
        ModbusBaseClient.__init__(
            self,
            framer,
            retries,
            self.comm_params,
            trace_packet,
            trace_pdu,
            trace_connect,
        )


class ModbusTcpClient(ModbusBaseSyncClient):
    """**ModbusTcpClient**.

    Fixed parameters:

    :param host: Host IP address or host name

    Optional parameters:

    :param framer: Framer name, default FramerType.SOCKET
    :param port: Port used for communication
    :param name: Set communication name, used in logging
    :param source_address: source address of client
    :param reconnect_delay: Not used in the sync client
    :param reconnect_delay_max: Not used in the sync client
    :param timeout: Timeout for connecting and receiving data, in seconds.
    :param retries: Max number of retries per request.
    :param trace_packet: Called with bytestream received/to be sent
    :param trace_pdu: Called with PDU received/to be sent
    :param trace_connect: Called when connected/disconnected

    .. tip::
        The trace methods allow to modify the datastream/pdu !

    Example::

        from pymodbus.client import ModbusTcpClient

        async def run():
            client = ModbusTcpClient("localhost")

            client.connect()
            ...
            client.close()

    Please refer to :ref:`Pymodbus internals` for advanced usage.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        host: str,
        *,
        framer: FramerType = FramerType.SOCKET,
        port: int = 502,
        name: str = "comm",
        source_address: tuple[str, int] | None = None,
        reconnect_delay: float = 0.1,
        reconnect_delay_max: float = 300,
        timeout: float = 3,
        retries: int = 3,
        trace_packet: Callable[[bool, bytes], bytes] | None = None,
        trace_pdu: Callable[[bool, ModbusPDU], ModbusPDU] | None = None,
        trace_connect: Callable[[bool], None] | None = None,
    ) -> None:
        """Initialize Modbus TCP Client."""
        if not hasattr(self,"comm_params"):
            if framer not in [FramerType.SOCKET, FramerType.RTU, FramerType.ASCII]:
                raise TypeError("Only FramerType SOCKET/RTU/ASCII allowed.")
            self.comm_params = CommParams(
                comm_type=CommType.TCP,
                host=host,
                port=port,
                comm_name=name,
                source_address=source_address,
                reconnect_delay=reconnect_delay,
                reconnect_delay_max=reconnect_delay_max,
                timeout_connect=timeout,
            )
        super().__init__(
            framer,
            retries,
            self.comm_params,
            trace_packet,
            trace_pdu,
            trace_connect,
        )
        self.socket: socket.socket | None = None

    @property
    def connected(self) -> bool:
        """Check if socket exists."""
        return self.socket is not None

    def connect(self):
        """Connect to the modbus tcp server."""
        if self.socket:
            return True
        try:
            self.socket = socket.create_connection(
                (self.comm_params.host, self.comm_params.port),
                timeout=self.comm_params.timeout_connect,
                source_address=self.comm_params.source_address,
            )
            Log.debug(
                "Connection to Modbus server established. Socket {}",
                self.socket.getsockname(),
            )
        except OSError as msg:
            Log.error(
                "Connection to ({}, {}) failed: {}",
                self.comm_params.host,
                self.comm_params.port,
                msg,
            )
            self.close()
        return self.socket is not None

    def close(self):
        """Close the underlying socket connection."""
        if self.socket:
            self.socket.close()
        self.socket = None

    def send(self, request, addr: tuple | None = None):
        """Send data on the underlying socket."""
        _ = addr
        if not self.socket:
            raise ConnectionException(str(self))
        if request:
            return self.socket.send(request)
        return 0

    def recv(self, size: int | None) -> bytes:
        """Read data from the underlying descriptor."""
        if not self.socket:
            raise ConnectionException(str(self))

        # socket.recv(size) waits until it gets some data from the host but
        # not necessarily the entire response that can be fragmented in
        # many packets.
        # To avoid split responses to be recognized as invalid
        # messages and to be discarded, loops socket.recv until full data
        # is received or timeout is expired.
        # If timeout expires returns the read data, also if its length is
        # less than the expected size.
        self.socket.setblocking(False)

        timeout = self.comm_params.timeout_connect or 0

        # If size isn't specified read up to 4096 bytes at a time.
        if size is None:
            recv_size = 4096
        else:
            recv_size = size

        data: list[bytes] = []
        data_length = 0
        time_ = time.time()
        end = time_ + timeout
        while recv_size > 0:
            try:
                ready = select.select([self.socket], [], [], end - time_)
            except ValueError:
                return self._handle_abrupt_socket_close(size, data, time.time() - time_)
            if ready[0]:
                try:
                    if (recv_data := self.socket.recv(recv_size)) == b"":
                        return self._handle_abrupt_socket_close(
                            size, data, time.time() - time_
                        )
                except SSLWantReadError:
                    continue
                data.append(recv_data)
                data_length += len(recv_data)
            time_ = time.time()

            # If size isn't specified continue to read until timeout expires.
            if not size:
                break
            # Timeout is reduced also if some data has been received in order
            # to avoid infinite loops when there isn't an expected response
            # size and the slave sends noisy data continuously.
            if time_ > end:
                break

            recv_size = size - data_length

        self.last_frame_end = round(time.time(), 6)
        return b"".join(data)

    def _handle_abrupt_socket_close(self, size: int | None, data: list[bytes], duration: float) -> bytes:
        """Handle unexpected socket close by remote end.

        Intended to be invoked after determining that the remote end
        has unexpectedly closed the connection, to clean up and handle
        the situation appropriately.

        :param size: The number of bytes that was attempted to read
        :param data: The actual data returned
        :param duration: Duration from the read was first attempted
               until it was determined that the remote closed the
               socket
        :return: The more than zero bytes read from the remote end
        :raises ConnectionException: If the remote end didn't send any
                 data at all before closing the connection.
        """
        self.close()
        size_txt = size if size else "unbounded read"
        readsize = f"read of {size_txt} bytes"
        msg = (
            f"{self}: Connection unexpectedly closed "
            f"{duration:.3f} seconds into {readsize}"
        )
        if data:
            result = b"".join(data)
            Log.warning(" after returning {} bytes: {} ", len(result), result)
            return result
        msg += " without response from slave before it closed connection"
        raise ConnectionException(msg)

    def is_socket_open(self) -> bool:
        """Check if socket is open."""
        return self.socket is not None

    def __repr__(self):
        """Return string representation."""
        return (
            f"<{self.__class__.__name__} at {hex(id(self))} socket={self.socket}, "
            f"ipaddr={self.comm_params.host}, port={self.comm_params.port}, timeout={self.comm_params.timeout_connect}>"
        )



"""**Server classes**."""

__all__ = [
    "ModbusSerialServer",
    "ModbusSimulatorServer",
    "ModbusTcpServer",
    "ModbusTlsServer",
    "ModbusUdpServer",
    "ServerAsyncStop",
    "ServerStop",
    "StartAsyncSerialServer",
    "StartAsyncTcpServer",
    "StartAsyncTlsServer",
    "StartAsyncUdpServer",
    "StartSerialServer",
    "StartTcpServer",
    "StartTlsServer",
    "StartUdpServer",
    "get_simulator_commandline",
]

from pymodbus.server.server import (
    ModbusSerialServer,
    ModbusTcpServer,
    ModbusTlsServer,
    ModbusUdpServer,
)
from pymodbus.server.simulator.http_server import ModbusSimulatorServer
from pymodbus.server.simulator.main import get_commandline as get_simulator_commandline
from pymodbus.server.startstop import (
    ServerAsyncStop,
    ServerStop,
    StartAsyncSerialServer,
    StartAsyncTcpServer,
    StartAsyncTlsServer,
    StartAsyncUdpServer,
    StartSerialServer,
    StartTcpServer,
    StartTlsServer,
    StartUdpServer,
)



# -*- coding: utf-8 -*-

"""package scinstr
author    Benoit Dubois
copyright FEMTO ENGINEERING, 2019-2025
license   GPL v3.0+
brief     API dedicated to handle the Keysight 532X0A counter serie.
"""

import logging
import socket
import usbtmc
from scinstr.cnt.constants import (
    CNT532X0A_PORT,
    CNT532X0A_TIMEOUT,
    CNT532X0A_VID,
    CNT532X0A_PID,
)


# =============================================================================
def isfloat(value):
    """Basic float test type function.
    """
    try:
        float(value)
        return True
    except ValueError:
        return False
    except TypeError:
        return False


# =============================================================================
class Cnt532x0aAbstract:
    """Abstract class to handling Counter532x0a digital multimeter device.
    Derived classes need to re-implement specific-protocol methods: connect(),
    close(), _write(), _read()...
    """

    def __del__(self):
        if self.is_connected:
            self.local()

    def connect(self):
        """Abstract protocol connect process. Derived classes must implement
        the connect process dedicated to the specific protocol used.
        :returns: None
        """
        raise NotImplementedError("Method not implemented by derived class")

    @property
    def is_connected(self):
        """Abstract connection check property.
        :returns: True if device connected else False (bool)
        """
        raise NotImplementedError("Method not implemented by derived class")

    def close(self):
        """Abstract protocol closing process. Derived classes must implement
        the closing process dedicated to the specific protocol used.
        :returns: None
        """
        raise NotImplementedError("Method not implemented by derived class")

    def _write(self, data) -> int:
        """Abstract protocol write process. Derived classes must implement
        the write process dedicated to the specific protocol used.
        :param data: data writes to device (str)
        :returns: None
        """
        raise NotImplementedError("Method not implemented by derived class")

    def _read(self) -> str | None:
        """Abstract protocol read process. Derived classes must implement
        the read process dedicated to the specific protocol used.
        :returns: Message reads from device (str)
        """
        raise NotImplementedError("Method not implemented by derived class")

    def write(self, data: str) -> int:
        """A basic write method: writes "data" to device.
        :param data: data writes to device (str)
        :returns: number of writed byte, -1 in case of error
        """
        try:
            retval = self._write(data)
        except Exception as ex:
            logging.error("Write error: %r", ex)
            return -1
        logging.debug(f"write: {data}")
        return retval

    def read(self) -> str | None:
        """A basic read method: read a message from device.
        :param length: length of message to read (int)
        :returns: Message reads from device, None if error (str)
        """
        try:
            retval = self._read()
        except Exception as ex:
            logging.error("Read error: %r", ex)
            return None
        logging.debug(f"read: {retval}")
        return retval

    def query(self, msg) -> None | str:
        """Read after write method.
        :param data: data writes to device (str)
        :param length: length of message to read (int)
        :returns: Message returned by device (str)
        """
        if self.write(msg) < 0:
            return None
        return self.read()

    def reset(self) -> bool:
        """Resets meter to its power-on state, sets all bits to zero in
        status byte register and all event registers and clear error queue.
        :returns: None
        """
        try:
            self._write("*RST")
            self._write("*CLS")
        except Exception as ex:
            logging.error(ex)
            return False
        logging.info("Counter reseted")
        return True

    def get_error(self) -> list[str] | None:
        """Return list of current error.
        :returns: list of current error (list of str)
        """
        errors = []
        while True:
            error = self.query("SYST:ERR?")
            if error is None or error == "":
                logging.error(f"Get error from device failed.")
                return None
            if "No error" in error:
                break
            errors.append(error)
        return errors

    def local(self):
        self.write("SYST:LOCK:REL")

    def data_read(self) -> float | None:
        """Takes a measurement the next time the trigger condition is met.
        After the measurement is taken, the reading is placed in the output
        buffer. "data_read" will not cause readings to be stored in the Meter’s
        internal memory.
        :returns: data read in buffer (float)
        """
        data = self.query("READ?")
        if data is None:
            logging.error(f"READ? returns None")
            return None
        try:
            data = float(data)
        except ValueError:
            logging.error(f"Invalid data {data}")
            return None
        return data


# =============================================================================
class Cnt532x0aEth(Cnt532x0aAbstract):
    """Class handling the 532x0a device through ethernet protocol."""

    def __init__(self, ip="", port=CNT532X0A_PORT, timeout=CNT532X0A_TIMEOUT):
        """The constructor.
        :param ip: IP address of device (str)
        :param port: socket port of device (int)
        :param timeout: socket timeout value in s (float)
        :returns: None
        """
        super().__init__()
        self._sock = None
        self.ip = ip
        self.port = port
        self._timeout = timeout

    def connect(self):
        """Specific ethernet connection process to Counter532x0a.
        :returns: True if connection success other False (Bool)
        """
        self._sock = socket.socket(
            socket.AF_INET, socket.SOCK_STREAM, socket.IPPROTO_TCP
        )
        self._sock.settimeout(self._timeout)
        try:
            self._sock.connect((self.ip, self.port))
        except ValueError as ex:
            logging.error("Connection parameters out of range: %r", ex)
            return False
        except socket.timeout:
            logging.error("Timeout on connection")
            return False
        except Exception as ex:
            logging.critical("Connection problem: %r", ex)
            return False
        logging.info("Connected to Counter532x0a")
        return True

    @property
    def is_connected(self):
        if self._sock is None:
            return False
        return True

    def close(self):
        """Specific ethernet closing process with Counter532x0a.
        :returns: None
        """
        super().close()
        try:
            self._sock.close()
        except Exception as ex:
            logging.error("%r", ex)
        self._sock = None
        logging.info("Connection to Counter532x0a closed")

    def _write(self, data):
        """Specific ethernet writing process.
        :param data: data writes to device (str)
        :returns: number of byte sent (int)
        """
        return self._sock.send((data + "\n").encode("utf8"))

    def _read(self, length):
        """Specific ethernet reading process.
        :param length: length of message to read (int)
        :returns: message reads from device (str)
        """
        return self._sock.recv(length).decode("utf-8").strip("\n")

    @property
    def timeout(self):
        """Gets timeout on socket operations.
        :returns: timeout value in second (float)
        """
        return self._sock.gettimeout()

    @timeout.setter
    def timeout(self, value):
        """Sets timeout on socket operations.
        :param value: timeout value in second (float)
        :returns: None
        """
        self._sock.settimeout(value)
        self._timeout = value


# =============================================================================
class Cnt532x0aUsb(Cnt532x0aAbstract):
    """Handle counter device through USB connection."""

    def __init__(
        self,
        vendor_id=CNT532X0A_VID,
        product_id=CNT532X0A_PID,
        timeout=CNT532X0A_TIMEOUT,
    ):
        self._dev = None
        self.vid = vendor_id
        self.pid = product_id
        self._timeout = timeout

    def connect(self):
        """Connect to the remote host.
        :returns: True if connection succeeded, False otherwise
        """
        logging.info("Connecting to counter")
        try:
            self._dev = usbtmc.Instrument(self.vid, self.pid)
        except usbtmc.usbtmc.UsbtmcException as ex:
            logging.error("Connection problem: %r", ex)
            return False
        self._dev.timeout = self._timeout
        logging.info("Connection --> Ok")
        return True

    @property
    def is_connected(self):
        if self._dev is None:
            return False
        return True

    def close(self):
        """Closes the underlying serial connection"""
        if self._dev is not None:
            try:
                self._dev.close()
            except Exception as ex:
                logging.error("Error when closing USB connection: %r", ex)
        self._dev = None
        logging.info("Connection to counter closed")

    def _write(self, data):
        """Specific USB writing process.
        :param data: data writes to device (str)
        :returns: number of bytes sent (int)
        """
        if not self.is_connected:
            self.connect()
        self._dev.write(data)
        return len(data)

    def _read(self):
        """Specific USB reading process.
        :returns: Message reads from device (str)
        """
        if not self.is_connected:
            self.connect()
        return self._dev.read()

    @property
    def timeout(self):
        """Gets timeout on socket operations.
        :returns: timeout value in second (float)
        """
        if self._dev is not None:
            return self._dev.timeout
        return None

    @timeout.setter
    def timeout(self, timeout):
        """Sets timeout on socket operations.
        :param timeout: timeout value in second (float)
        :returns: None
        """
        if self._dev is not None:
            self._dev.timeout = timeout
            self._timeout = timeout


# =============================================================================
def check_counter532x0a():
    """Checks the Counter532x0axx class: connect to the counter, configure
    a basic measurement then collects and print data to standard output.
    """
    from datetime import datetime

    date_fmt = "%d/%m/%Y %H:%M:%S"
    log_format = (
        "%(asctime)s %(levelname) -8s %(filename)s "
        + " %(funcName)s (%(lineno)d): %(message)s"
    )
    logging.basicConfig(level=logging.INFO, datefmt=date_fmt, format=log_format)

    counter = Cnt532x0aUsb(0x0957, 0x1707, timeout=2.8)
    # counter = Cnt532x0aEth(ip="192.168.0.20", port=5025, timeout=1.5)
    if counter.connect() is not True:
        print("Connection failed")
        return
    counter.reset()

    print("IDN:", counter.idn())
    counter.write("CONF:FREQ 100.0E6")
    counter.write("TRIG:SOUR IMM")
    counter.write("TRIG:SLOP POS")
    counter.write("SENS:FREQ:GATE:TIME 1.0")
    counter.write("SENS:FREQ:GATE:SOUR TIME")
    print("Error config?:", counter.get_error())

    try:
        while True:
            value = counter.data_read()
            now = datetime.utcnow()
            if value is None or value == "":
                print("# No data @", now)
            else:
                print(now, value)
    except KeyboardInterrupt:
        counter.write("ABORT")
    except Exception as er:
        logging.error("# Exception during acquisition: %r", er)

    counter.close()


# =============================================================================
if __name__ == "__main__":
    check_counter532x0a()



# -*- coding: utf-8 -*-

"""package scinstr.daq.labjack
author    Benoit Dubois
copyright FEMTO Engineering, 2020
licence   GPL 3.0+
brief     Handle Labjack T7(-Pro) device (DAQ board).
"""

import logging
from collections import namedtuple
import scinstr.daq.labjack.convert_data as cd


AIN_ADDR = [0 + addr for addr in range(0, 27, 2)]
AIN_BINARY_ADDR = [50000 + addr*2 for addr in range(0, 14)]
#
CURRENT_SOURCE_10UA_CAL_VALUE_ADDR = 1900
CURRENT_SOURCE_200UA_CAL_VALUE_ADDR = 1902
#
FIO_ADDR = [2000 + addr for addr in range(0, 8)]
EIO_ADDR = [2008 + addr for addr in range(0, 8)]
CIO_ADDR = [2016 + addr for addr in range(0, 8)]
MIO_ADDR = [2020 + addr for addr in range(0, 8)]
FIO_DIRECTION_ADDR = 2600  # 0=Input and 1=Output
EIO_DIRECTION_ADDR = 2601  # 0=Input and 1=Output
CIO_DIRECTION_ADDR = 2602  # 0=Input and 1=Output
MIO_DIRECTION_ADDR = 2603  # 0=Input and 1=Output
FIO_STATE_ADDR = 2500
EIO_STATE_ADDR = 2501
CIO_STATE_ADDR = 2502
MIO_STATE_ADDR = 2503
#
DIO_STATE_ADDR = 2800
DIO_DIRECTION_ADDR = 2850  # 0=Input and 1=Output
DIO_INHIBIT_ADDR = 2900  # 0=Affected (Default),1=Ignored
#
DIO_EF_READ_A_ADDR = [3000 + addr*2 for addr in range(0, 23)]
DIO_EF_READ_A_AND_RESET_ADDR = [3100 + addr*2 for addr in range(0, 23)]
DIO_EF_READ_A_F_ADDR = [3500 + addr*2 for addr in range(0, 23)]
DIO_EF_READ_A_F_AND_RESET_ADDR = [3600 + addr*2 for addr in range(0, 23)]
DIO_EF_ENABLE_ADDR = [44000 + addr*2 for addr in range(0, 23)]
DIO_EF_INDEX_ADDR = [44100 + addr*2 for addr in range(0, 23)]
DIO_EF_OPTIONS_ADDR = [44200 + addr*2 for addr in range(0, 23)]
#
DIO_EF_CLOCK_ENABLE_ADDR = [44900 + addr*10 for addr in range(0, 8)]
DIO_EF_CLOCK_DIVISOR_ADDR = [44901 + addr*10 for addr in range(0, 8)]
DIO_EF_CLOCK_OPTIONS_ADDR = [44902 + addr*10 for addr in range(0, 8)]
DIO_EF_CLOCK_ROLL_VALUE_ADDR = [44904 + addr*10 for addr in range(0, 8)]
DIO_EF_CLOCK_COUNT_ADDR = [44908 + addr*10 for addr in range(0, 8)]
#
STREAM_SCANRATE_HZ = 4002
STREAM_NUM_ADDRESSES = 4004
STREAM_SAMPLES_PER_PACKET = 4006
STREAM_SETTLING_US = 4008
STREAM_RESOLUTION_INDEX = 4010
STREAM_BUFFER_SIZE_BYTES = 4012
# STREAM_CLOCK_SOURCE = 4014
# STREAM_OPTIONS = 4014
STREAM_AUTO_TARGET = 4016
STREAM_DATATYPE = 4018
STREAM_NUM_SCANS = 4020
STREAM_ENABLE = 4990
STREAM_SCANLIST_ADDRESS = [4100 + addr for addr in range(0, 128, 2)]
#
AIN_RANGE_ADDR = [40000 + addr for addr in range(0, 27, 2)]
AIN_NEGATIVE_CH_ADDR = [41000 + addr for addr in range(0, 14)]
AIN_RESOLUTION_INDEX_ADDR = [41500 + addr for addr in range(0, 14)]
AIN_SETTLING_US_ADDR = [42000 + addr for addr in range(0, 27, 2)]
#
POWER_ETHERNET_ADDR = 48003
POWER_WIFI_ADDR = 48004
POWER_LED_ADDR = 48006
POWER_ETHERNET_DEFAULT_ADDR = 48053
POWER_WIFI_DEFAULT_ADDR = 48054
#
LED_COMM_ADDR =	2990
LED_STATUS_ADDR = 2991
#
IO_CONFIG_SET_DEFAULT_TO_CURRENT = 49002
#
TEST = 55100
TEST_RESULT = 1122867
TEST_UINT16 = 55110
TEST_UINT16_RESULT = 17
TEST_UINT32 = 55120
TEST_UINT32_RESULT = 1122867
TEST_INT32 = 55122
TEST_INT32_RESULT = -2003195205
TEST_FLOAT32 = 55124
TEST_FLOAT32_RESULT = -9999.0
#
PRODUCT_ID = 60000
#
WATCHDOG_ENABLE_DEFAULT = 61600
WATCHDOG_TIMEOUT_S_DEFAULT = 61604
WATCHDOG_RESET_ENABLE_DEFAULT = 61620
#
SYSTEM_REBOOT = 61998

# Define a new type used for storing device configurations
DTuple = namedtuple('DTuple', ['mnemo', 'caption', 'list'])


# T7(-Pro) device capabilities namespace
# =============================================================================
RESO_T7PRO = [DTuple("0", "Default 9 (19.6 bits, gain=1)", None),
              DTuple("1", "1 (16 bits, gain=1)", None),
              DTuple("2", "2", None),
              DTuple("3", "3", None),
              DTuple("4", "4", None),
              DTuple("5", "5", None),
              DTuple("6", "6", None),
              DTuple("7", "7", None),
              DTuple("8", "8", None),
              DTuple("9", "9 (19.6 bits, gain=1)", None),
              DTuple("10", "10", None),
              DTuple("11", "11", None),
              DTuple("12", "12 (21.8 bits, gain=1)", None)]

RESO_T7 = [DTuple("0", "Default 8 (19.1 bits, gain=1)", None),
           DTuple("1", "1 (16 bits, gain=1)", None),
           DTuple("2", "2", None),
           DTuple("3", "3", None),
           DTuple("4", "4", None),
           DTuple("5", "5", None),
           DTuple("6", "6", None),
           DTuple("7", "7", None),
           DTuple("8", "8 (19.1 bits, gain=1)", None)]

VOLT_RANGE = [DTuple("0.01", "10 mV", None),
              DTuple("0.1", "100 mV", None),
              DTuple("1", "1 V", None),
              DTuple("10", "10 V", None)]

SETTLING = [DTuple("0", "auto", None),
            DTuple("100", "100 us", None),
            DTuple("1000", "1 ms", None),
            DTuple("2000", "2 ms", None),
            DTuple("10000", "10 ms", None),
            DTuple("20000", "20 ms", None),
            DTuple("100000", "100 ms", None),
            DTuple("200000", "200 ms", None)]

CUR_SRC = [DTuple("0", "Disabled", None),
           DTuple("1", "10 uA", None),
           DTuple("2", "200 uA", None)]

FUNCTION = [DTuple("Voltage", "DC Voltage", VOLT_RANGE)]
# , DTuple("Pt100", "2-wire Pt100", None)] ##CUR_SRC[0])]


# =============================================================================
class T7():
    """Base class to handle T7(-Pro) board.
    """

    def __init__(self):
        """Constructor.
        :returns: None
        """
        self._client = None

    def connect(self):
        """ Connect to the remote host
        :returns: True if connection succeeded, False otherwise
        """
        raise NotImplementedError("Method not implemented by derived class")

    def close(self):
        """ Closes the underlying socket connection
        """
        raise NotImplementedError("Method not implemented by derived class")

    def reboot(self):
        """Reboot device.
        :returns: None
        """
        # Reboot after 50 ms
        self._client.write_registers(SYSTEM_REBOOT,
                                     cd.uint32_to_data(0x4C4A0001))

    def check_test_reg(self):
        """Use test registers to test device.
        :returns: True if test pass else False (bool)
        """
        rr = self._client.read_input_registers(TEST, 2)
        test = cd.data_to_uint32(rr.registers)
        if test != TEST_RESULT:
            return False
        return True

    def set_current_config_as_default(self):
        """Set current device configuration as new default (ie after
        reboot/power-up) values.
        :returns: None
        """
        self._client.write_registers(IO_CONFIG_SET_DEFAULT_TO_CURRENT,
                                     cd.uint32_to_data(1))

    def set_watchdog(self, timeout=60):
        """Configure watchdog to reset the device if it does not receive any
        communication for 'timeout' seconds. Usualy, the method is used in
        conjonction with 'set_current_config_as_default()' method.
        :param timeout: watchdog timer value (int)
        :returns: None
        """
        assert timeout > 1, \
            "Watchdog timeout value (%r s) too small." % timeout
        self._client.write_registers(WATCHDOG_ENABLE_DEFAULT,
                                     cd.uint32_to_data(0))
        self._client.write_registers(WATCHDOG_TIMEOUT_S_DEFAULT,
                                     cd.uint32_to_data(timeout))
        self._client.write_registers(WATCHDOG_RESET_ENABLE_DEFAULT,
                                     cd.uint32_to_data(1))
        self._client.write_registers(WATCHDOG_ENABLE_DEFAULT,
                                     cd.uint32_to_data(1))

    def read_float32_register(self, register):
        """Read a register of float 32 bits type.
        :returns: the register value (float)
        """
        rr = self._client.read_input_registers(register, 2)
        return cd.data_to_float32(rr.registers)

    def write_float32_register(self, register, value):
        """Write a register of float 32 bits type.
        :param register: address to write (int)
        :param value: value to write (float)
        :returns: None
        """
        self._client.write_registers(register,
                                     cd.float32_to_data(value))

    def read_uint16_register(self, register):
        """Read a register of unsigned integer 16 bits type.
        :returns: the register value (int)
        """
        rr = self._client.read_input_registers(register, 2)
        return cd.data_to_uint16(rr.registers)

    def write_uint16_register(self, register, value):
        """Write a register of unsigned integer 16 bits type.
        :param register: address to write (int)
        :param value: value to write (int)
        :returns: None
        """
        self._client.write_registers(register,
                                     cd.uint16_to_data(value))

    def read_uint32_register(self, register):
        """Read a register of unsigned integer 32 bits type.
        :returns: the register value (int)
        """
        rr = self._client.read_input_registers(register, 2)
        return cd.data_to_uint32(rr.registers)

    def write_uint32_register(self, register, value):
        """Write a register of unsigned integer 32 bits type.
        :param register: address to write (int)
        :param value: value to write (int)
        :returns: None
        """
        self._client.write_registers(register,
                                     cd.uint32_to_data(value))

    def read_int32_register(self, register):
        """Read a register of integer 32 bits type.
        :returns: the register value (int)
        """
        rr = self._client.read_input_registers(register, 2)
        return cd.data_to_int32(rr.registers)

    def write_int32_register(self, register, value):
        """Write a register of integer 32 bits type.
        :param register: address to write (int)
        :param value: value to write (int)
        :returns: None
        """
        self._client.write_registers(register,
                                     cd.int32_to_data(value))

    def get_id(self):
        """Return product id of device.
        :returns: product id of device (str)
        """
        rr = self._client.read_input_registers(PRODUCT_ID, 2)
        return cd.data_to_float32(rr.registers)

    def get_current_source_10ua_value(self, r_calibrate=None, v_ain=None):
        """Return the intensity value of the 10 uA current source.
        If a calibrate resistor is used, return the actual value computed from
        I/V value, else return the factory calibration value.
        :returns: 10 uA current source value (float)
        """
        if r_calibrate is None or v_ain is None:
            rr = self._client.read_input_registers(
                CURRENT_SOURCE_10UA_CAL_VALUE_ADDR, 2)
            current = cd.data_to_float32(rr.registers)
        else:
            current = v_ain / r_calibrate
        return current

    def get_current_source_200ua_value(self, r_calibrate=None, v_ain=None):
        """Return the intensity value of the 200 uA current source.
        If a calibrate resistor is used, return the actual value computed from
        I/V value, else return the factory calibration value.
        :returns: 200 uA current source value (float)
        """
        if r_calibrate is None or v_ain is None:
            rr = self._client.read_input_registers(
                CURRENT_SOURCE_200UA_CAL_VALUE_ADDR, 2)
            current = cd.data_to_float32(rr.registers)
        else:
            current = v_ain / r_calibrate
        return current

    def get_ains_voltage(self, ains):
        """Read analog input values.
        :param ains: list of analog inputs (list of int)
        :returns: list of analog input value(s) (list of float)
        """
        vains = []
        for ain in ains:
            assert 0 <= ain <= 13, "channel index out of range: %r" % ain
            rr = self._client.read_input_registers(AIN_ADDR[ain], 2)
            vains.append(cd.data_to_float32(rr.registers))
        return vains

    def set_ains_range(self, ains, ranges):
        """Set analog input range.
        :param ains: list of analog inputs (list of int)
        :param ranges: list of analog input range (list of float)
        :returns: None
        """
        assert len(ains) == len(ranges), \
            "parameters must have the same size"
        for ain, _range in zip(ains, ranges):
            self._client.write_registers(AIN_RANGE_ADDR[ain],
                                         cd.float32_to_data(_range))

    def set_ains_negative_ch(self, ains, negative_chs):
        """Set negative channel to be used for each positive channel.
        :param ains: list of analog positive channel (list of int)
        :param negative_chs: list of analog negative channel (list of int)
        :returns: None
        """
        assert len(ains) == len(negative_chs), \
            "parameters must have the same size"
        for ain, negative_ch in zip(ains, negative_chs):
            assert ain % 2 == 0, \
              "Even negative_ch registers must not be writed: %r" % ain
            self._client.write_register(AIN_NEGATIVE_CH_ADDR[ain],
                                        cd.uint16_to_data(negative_ch))

    def set_ains_resolution(self, ains, resolutions):
        """Set analog input resolution.
        :param ains: list of analog inputs (list of int)
        :param resolutions: list of analog input resolution (list of int)
        :returns: None
        """
        assert len(ains) == len(resolutions), \
            "parameters must have the same size"
        for ain, resolution in zip(ains, resolutions):
            assert 0 <= ain <= 13, "Analog input out of range: %r" % ain
            assert 0 <= resolution <= 12, \
                "Resolution out of range: %r" % resolution
            self._client.write_register(cd.uint16_to_data(resolution),
                                        AIN_RESOLUTION_INDEX_ADDR[ain])

    def set_ains_settling(self, ains, settlings):
        """Set analog input settling time.
        :param ains: list of analog inputs (list of int)
        :param settlings: list of analog input settling time (list of float)
        :returns: None
        """
        assert len(ains) == len(settlings), \
            "parameters must have the same size"
        for ain, settling in zip(ains, settlings):
            assert 0 <= ain <= 13, "Analog input out of range: %r" % ain
            assert 0 <= settling <= 3.402823 * 10**38, \
                "Settling out of range: %r" % settling
            self._client.write_registers(AIN_SETTLING_US_ADDR[ain],
                                         cd.float32_to_data(settling))

    def enable_high_speed_counter(self, num):
        """Enable high speed counter.
        :param num: number of counter (0 to 3) to enable (int)
        :returns: None
        """
        assert 0 <= num <= 3, "Counter index out of range: %r" % num
        if num == 0:
            self.write_uint16_register(DIO_EF_CLOCK_ENABLE_ADDR[0], 0)
            self.write_uint16_register(DIO_EF_CLOCK_ENABLE_ADDR[2], 0)
        if num == 1:
            self.write_uint16_register(DIO_EF_CLOCK_ENABLE_ADDR[0], 0)
            self.write_uint16_register(DIO_EF_CLOCK_ENABLE_ADDR[1], 0)
        if num == 3:
            # Disable stream mode (not implemented yet)
            logging.error("Configuration of High Speed counter 3 need to "
                          "disable stream mode before (not implemented).")
        self.write_uint32_register(DIO_EF_ENABLE_ADDR[num+16], 0)
        self.write_uint32_register(DIO_EF_INDEX_ADDR[num+16], 7)
        self.write_uint32_register(DIO_EF_ENABLE_ADDR[num+16], 1)

    def get_high_speed_counter_frequency(self, num):
        """Get frequency measured by high speed counter.
        :param num: number of counter (0 to 3) to enable (int)
        :returns: frequency measured by high speed counter (float)
        """
        assert 0 <= num <= 3, "Counter index out of range: %r" % num
        return self.read_uint32_register(DIO_EF_READ_A_AND_RESET_ADDR[num+16])

    def enable_fios(self, num, direction):
        """Enable FIO digital I/O and set direction.
        :param num: list of I/O to enable (list of int)
        :param direction: list of I/O direction (list of int)
        :returns: None
        """
        msg = 0
        for idx, n in enumerate(num):
            assert 0 <= n <= 7, "FIO index out of range: %r" % num
            msg += direction[idx] * 2**n
        self.write_uint16_register(FIO_DIRECTION_ADDR, msg)

    def set_fios_state(self, num, state):
        """Set state of FIO digital output.
        :param num: list of I/O to set state (list of int)
        :param state: list of state (list of int)
        :returns: None
        """
        msg = 0
        for idx, n in enumerate(num):
            assert 0 <= n <= 7, "FIO index out of range: %r" % num
            msg += state[idx] * 2**n
        self.write_uint16_register(FIO_STATE_ADDR, msg)

    def get_fios_state(self, num):
        """Get state of FIO digital output.
        :param num: list of I/O to get state (list of int)
        :returns: list of state (list of int)
        """
        fios_state = self.read_uint16_register(FIO_STATE_ADDR)
        state_list = []
        for idx, n in enumerate(num):
            assert 0 <= n <= 7, "FIO index out of range: %r" % num
            state_list.append((x >> n) & 1)
        return state_list



"""
    pygments.util
    ~~~~~~~~~~~~~

    Utility functions.

    :copyright: Copyright 2006-2024 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

import re
from io import TextIOWrapper


split_path_re = re.compile(r'[/\\ ]')
doctype_lookup_re = re.compile(r'''
    <!DOCTYPE\s+(
     [a-zA-Z_][a-zA-Z0-9]*
     (?: \s+      # optional in HTML5
     [a-zA-Z_][a-zA-Z0-9]*\s+
     "[^"]*")?
     )
     [^>]*>
''', re.DOTALL | re.MULTILINE | re.VERBOSE)
tag_re = re.compile(r'<(.+?)(\s.*?)?>.*?</.+?>',
                    re.IGNORECASE | re.DOTALL | re.MULTILINE)
xml_decl_re = re.compile(r'\s*<\?xml[^>]*\?>', re.I)


class ClassNotFound(ValueError):
    """Raised if one of the lookup functions didn't find a matching class."""


class OptionError(Exception):
    """
    This exception will be raised by all option processing functions if
    the type or value of the argument is not correct.
    """

def get_choice_opt(options, optname, allowed, default=None, normcase=False):
    """
    If the key `optname` from the dictionary is not in the sequence
    `allowed`, raise an error, otherwise return it.
    """
    string = options.get(optname, default)
    if normcase:
        string = string.lower()
    if string not in allowed:
        raise OptionError('Value for option {} must be one of {}'.format(optname, ', '.join(map(str, allowed))))
    return string


def get_bool_opt(options, optname, default=None):
    """
    Intuitively, this is `options.get(optname, default)`, but restricted to
    Boolean value. The Booleans can be represented as string, in order to accept
    Boolean value from the command line arguments. If the key `optname` is
    present in the dictionary `options` and is not associated with a Boolean,
    raise an `OptionError`. If it is absent, `default` is returned instead.

    The valid string values for ``True`` are ``1``, ``yes``, ``true`` and
    ``on``, the ones for ``False`` are ``0``, ``no``, ``false`` and ``off``
    (matched case-insensitively).
    """
    string = options.get(optname, default)
    if isinstance(string, bool):
        return string
    elif isinstance(string, int):
        return bool(string)
    elif not isinstance(string, str):
        raise OptionError(f'Invalid type {string!r} for option {optname}; use '
                          '1/0, yes/no, true/false, on/off')
    elif string.lower() in ('1', 'yes', 'true', 'on'):
        return True
    elif string.lower() in ('0', 'no', 'false', 'off'):
        return False
    else:
        raise OptionError(f'Invalid value {string!r} for option {optname}; use '
                          '1/0, yes/no, true/false, on/off')


def get_int_opt(options, optname, default=None):
    """As :func:`get_bool_opt`, but interpret the value as an integer."""
    string = options.get(optname, default)
    try:
        return int(string)
    except TypeError:
        raise OptionError(f'Invalid type {string!r} for option {optname}; you '
                          'must give an integer value')
    except ValueError:
        raise OptionError(f'Invalid value {string!r} for option {optname}; you '
                          'must give an integer value')

def get_list_opt(options, optname, default=None):
    """
    If the key `optname` from the dictionary `options` is a string,
    split it at whitespace and return it. If it is already a list
    or a tuple, it is returned as a list.
    """
    val = options.get(optname, default)
    if isinstance(val, str):
        return val.split()
    elif isinstance(val, (list, tuple)):
        return list(val)
    else:
        raise OptionError(f'Invalid type {val!r} for option {optname}; you '
                          'must give a list value')


def docstring_headline(obj):
    if not obj.__doc__:
        return ''
    res = []
    for line in obj.__doc__.strip().splitlines():
        if line.strip():
            res.append(" " + line.strip())
        else:
            break
    return ''.join(res).lstrip()


def make_analysator(f):
    """Return a static text analyser function that returns float values."""
    def text_analyse(text):
        try:
            rv = f(text)
        except Exception:
            return 0.0
        if not rv:
            return 0.0
        try:
            return min(1.0, max(0.0, float(rv)))
        except (ValueError, TypeError):
            return 0.0
    text_analyse.__doc__ = f.__doc__
    return staticmethod(text_analyse)


def shebang_matches(text, regex):
    r"""Check if the given regular expression matches the last part of the
    shebang if one exists.

        >>> from pygments.util import shebang_matches
        >>> shebang_matches('#!/usr/bin/env python', r'python(2\.\d)?')
        True
        >>> shebang_matches('#!/usr/bin/python2.4', r'python(2\.\d)?')
        True
        >>> shebang_matches('#!/usr/bin/python-ruby', r'python(2\.\d)?')
        False
        >>> shebang_matches('#!/usr/bin/python/ruby', r'python(2\.\d)?')
        False
        >>> shebang_matches('#!/usr/bin/startsomethingwith python',
        ...                 r'python(2\.\d)?')
        True

    It also checks for common windows executable file extensions::

        >>> shebang_matches('#!C:\\Python2.4\\Python.exe', r'python(2\.\d)?')
        True

    Parameters (``'-f'`` or ``'--foo'`` are ignored so ``'perl'`` does
    the same as ``'perl -e'``)

    Note that this method automatically searches the whole string (eg:
    the regular expression is wrapped in ``'^$'``)
    """
    index = text.find('\n')
    if index >= 0:
        first_line = text[:index].lower()
    else:
        first_line = text.lower()
    if first_line.startswith('#!'):
        try:
            found = [x for x in split_path_re.split(first_line[2:].strip())
                     if x and not x.startswith('-')][-1]
        except IndexError:
            return False
        regex = re.compile(rf'^{regex}(\.(exe|cmd|bat|bin))?$', re.IGNORECASE)
        if regex.search(found) is not None:
            return True
    return False


def doctype_matches(text, regex):
    """Check if the doctype matches a regular expression (if present).

    Note that this method only checks the first part of a DOCTYPE.
    eg: 'html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"'
    """
    m = doctype_lookup_re.search(text)
    if m is None:
        return False
    doctype = m.group(1)
    return re.compile(regex, re.I).match(doctype.strip()) is not None


def html_doctype_matches(text):
    """Check if the file looks like it has a html doctype."""
    return doctype_matches(text, r'html')


_looks_like_xml_cache = {}


def looks_like_xml(text):
    """Check if a doctype exists or if we have some tags."""
    if xml_decl_re.match(text):
        return True
    key = hash(text)
    try:
        return _looks_like_xml_cache[key]
    except KeyError:
        m = doctype_lookup_re.search(text)
        if m is not None:
            return True
        rv = tag_re.search(text[:1000]) is not None
        _looks_like_xml_cache[key] = rv
        return rv


def surrogatepair(c):
    """Given a unicode character code with length greater than 16 bits,
    return the two 16 bit surrogate pair.
    """
    # From example D28 of:
    # http://www.unicode.org/book/ch03.pdf
    return (0xd7c0 + (c >> 10), (0xdc00 + (c & 0x3ff)))


def format_lines(var_name, seq, raw=False, indent_level=0):
    """Formats a sequence of strings for output."""
    lines = []
    base_indent = ' ' * indent_level * 4
    inner_indent = ' ' * (indent_level + 1) * 4
    lines.append(base_indent + var_name + ' = (')
    if raw:
        # These should be preformatted reprs of, say, tuples.
        for i in seq:
            lines.append(inner_indent + i + ',')
    else:
        for i in seq:
            # Force use of single quotes
            r = repr(i + '"')
            lines.append(inner_indent + r[:-2] + r[-1] + ',')
    lines.append(base_indent + ')')
    return '\n'.join(lines)


def duplicates_removed(it, already_seen=()):
    """
    Returns a list with duplicates removed from the iterable `it`.

    Order is preserved.
    """
    lst = []
    seen = set()
    for i in it:
        if i in seen or i in already_seen:
            continue
        lst.append(i)
        seen.add(i)
    return lst


class Future:
    """Generic class to defer some work.

    Handled specially in RegexLexerMeta, to support regex string construction at
    first use.
    """
    def get(self):
        raise NotImplementedError


def guess_decode(text):
    """Decode *text* with guessed encoding.

    First try UTF-8; this should fail for non-UTF-8 encodings.
    Then try the preferred locale encoding.
    Fall back to latin-1, which always works.
    """
    try:
        text = text.decode('utf-8')
        return text, 'utf-8'
    except UnicodeDecodeError:
        try:
            import locale
            prefencoding = locale.getpreferredencoding()
            text = text.decode()
            return text, prefencoding
        except (UnicodeDecodeError, LookupError):
            text = text.decode('latin1')
            return text, 'latin1'


def guess_decode_from_terminal(text, term):
    """Decode *text* coming from terminal *term*.

    First try the terminal encoding, if given.
    Then try UTF-8.  Then try the preferred locale encoding.
    Fall back to latin-1, which always works.
    """
    if getattr(term, 'encoding', None):
        try:
            text = text.decode(term.encoding)
        except UnicodeDecodeError:
            pass
        else:
            return text, term.encoding
    return guess_decode(text)


def terminal_encoding(term):
    """Return our best guess of encoding for the given *term*."""
    if getattr(term, 'encoding', None):
        return term.encoding
    import locale
    return locale.getpreferredencoding()


class UnclosingTextIOWrapper(TextIOWrapper):
    # Don't close underlying buffer on destruction.
    def close(self):
        self.flush()



"""
    pygments.styles
    ~~~~~~~~~~~~~~~

    Contains built-in styles.

    :copyright: Copyright 2006-2024 by the Pygments team, see AUTHORS.
    :license: BSD, see LICENSE for details.
"""

from pip._vendor.pygments.plugin import find_plugin_styles
from pip._vendor.pygments.util import ClassNotFound
from pip._vendor.pygments.styles._mapping import STYLES

#: A dictionary of built-in styles, mapping style names to
#: ``'submodule::classname'`` strings.
#: This list is deprecated. Use `pygments.styles.STYLES` instead
STYLE_MAP = {v[1]: v[0].split('.')[-1] + '::' + k for k, v in STYLES.items()}

#: Internal reverse mapping to make `get_style_by_name` more efficient
_STYLE_NAME_TO_MODULE_MAP = {v[1]: (v[0], k) for k, v in STYLES.items()}


def get_style_by_name(name):
    """
    Return a style class by its short name. The names of the builtin styles
    are listed in :data:`pygments.styles.STYLE_MAP`.

    Will raise :exc:`pygments.util.ClassNotFound` if no style of that name is
    found.
    """
    if name in _STYLE_NAME_TO_MODULE_MAP:
        mod, cls = _STYLE_NAME_TO_MODULE_MAP[name]
        builtin = "yes"
    else:
        for found_name, style in find_plugin_styles():
            if name == found_name:
                return style
        # perhaps it got dropped into our styles package
        builtin = ""
        mod = 'pygments.styles.' + name
        cls = name.title() + "Style"

    try:
        mod = __import__(mod, None, None, [cls])
    except ImportError:
        raise ClassNotFound(f"Could not find style module {mod!r}" +
                            (builtin and ", though it should be builtin")
                            + ".")
    try:
        return getattr(mod, cls)
    except AttributeError:
        raise ClassNotFound(f"Could not find style class {cls!r} in style module.")


def get_all_styles():
    """Return a generator for all styles by name, both builtin and plugin."""
    for v in STYLES.values():
        yield v[1]
    for name, _ in find_plugin_styles():
        yield name



"""Constants For Modbus Server/Client.

This is the single location for storing default
values for the servers and clients.
"""
import enum


INTERNAL_ERROR = "Pymodbus internal error"


class ModbusStatus(int, enum.Enum):
    """These represent various status codes in the modbus protocol.

    .. attribute:: WAITING

       This indicates that a modbus device is currently
       waiting for a given request to finish some running task.

    .. attribute:: READY

       This indicates that a modbus device is currently
       free to perform the next request task.

    .. attribute:: ON

       This indicates that the given modbus entity is on

    .. attribute:: OFF

       This indicates that the given modbus entity is off

    .. attribute:: SLAVE_ON

       This indicates that the given modbus slave is running

    .. attribute:: SLAVE_OFF

       This indicates that the given modbus slave is not running
    """

    WAITING = 0xFFFF
    READY = 0x0000
    ON = 0xFF00
    OFF = 0x0000
    SLAVE_ON = 0xFF
    SLAVE_OFF = 0x00


class Endian(str, enum.Enum):
    """An enumeration representing the various byte endianness.

    .. attribute:: AUTO

       This indicates that the byte order is chosen by the
       current native environment.

    .. attribute:: BIG

       This indicates that the bytes are in big endian format

    .. attribute:: LITTLE

       This indicates that the bytes are in little endian format

    .. note:: I am simply borrowing the format strings from the
       python struct module for my convenience.
    """

    AUTO = "@"
    BIG = ">"
    LITTLE = "<"


class ModbusPlusOperation(int, enum.Enum):
    """Represents the type of modbus plus request.

    .. attribute:: GET_STATISTICS

       Operation requesting that the current modbus plus statistics
       be returned in the response.

    .. attribute:: CLEAR_STATISTICS

       Operation requesting that the current modbus plus statistics
       be cleared and not returned in the response.
    """

    GET_STATISTICS = 0x0003
    CLEAR_STATISTICS = 0x0004


class DeviceInformation(int, enum.Enum):
    """Represents what type of device information to read.

    .. attribute:: BASIC

       This is the basic (required) device information to be returned.
       This includes VendorName, ProductCode, and MajorMinorRevision
       code.

    .. attribute:: REGULAR

       In addition to basic data objects, the device provides additional
       and optional identification and description data objects. All of
       the objects of this category are defined in the standard but their
       implementation is optional.

    .. attribute:: EXTENDED

       In addition to regular data objects, the device provides additional
       and optional identification and description private data about the
       physical device itself. All of these data are device dependent.

    .. attribute:: SPECIFIC

       Request to return a single data object.
    """

    BASIC = 0x01
    REGULAR = 0x02
    EXTENDED = 0x03
    SPECIFIC = 0x04

    def __str__(self):
        """Override to force int representation for enum members."""
        return str(int(self))


class MoreData(int, enum.Enum):
    """Represents the more follows condition.

    .. attribute:: NOTHING

       This indicates that no more objects are going to be returned.

    .. attribute:: KEEP_READING

       This indicates that there are more objects to be returned.
    """

    NOTHING = 0x00
    KEEP_READING = 0xFF




# -*- coding: utf-8 -*-

# Default Agilent/Keysight 532x0A counter parameter
CNT532X0A_ID = "Technologies,532"
CNT532X0A_PORT = 5025
CNT532X0A_TIMEOUT = 3.0
CNT532X0A_VID = 0x0957
CNT532X0A_PID = 0x1707



"""ModbusProtocol layer.

Contains pure transport methods needed to
- connect/listen,
- send/receive
- close/abort connections
for unix socket, tcp, tls and serial communications as well as a special
null modem option.

Contains high level methods like reconnect.

All transport differences are handled in transport, providing a unified
interface to upper layers.

Host/Port/SourceAddress explanation:
- SourceAddress (host, port):
- server (host, port): Listen on host:port
- server serial (comm_port, _): comm_port is device string
- client (host, port): Bind host:port to interface
- client serial: not used
- Host
- server: not used
- client: remote host to connect to (as host:port)
- client serial: host is comm_port device string
- Port
- server: not used
- client: remote port to connect to (as host:port)
- client serial: no used

Pyserial allow the comm_port to be a socket e.g. "socket://localhost:502",
this allows serial clients to connect to a tcp server with RTU framer.

Pymodbus allows this format for both server and client.
For clients the string is passed to pyserial,
but for servers it is used to start a modbus tcp server.
This allows for serial testing, without a serial cable.

Pymodbus offers nullmodem for clients/servers running in the same process
if <host> is set to NULLMODEM_HOST it will be automatically invoked.
This allows testing without actual network traffic and is a lot faster.

Class NullModem is a asyncio transport class,
that replaces the socket class or pyserial.

The class is designed to take care of differences between the different
transport mediums, and provide a neutral interface for the upper layers.
It basically provides a pipe, without caring about the actual data content.
"""
from __future__ import annotations

import asyncio
import dataclasses
import ssl
from abc import abstractmethod
from collections.abc import Callable, Coroutine
from contextlib import suppress
from enum import Enum
from functools import partial
from typing import Any

from pymodbus.logging import Log
from pymodbus.transport.serialtransport import create_serial_connection


NULLMODEM_HOST = "__pymodbus_nullmodem"


class CommType(int, Enum):
    """Type of transport."""

    TCP = 1
    TLS = 2
    UDP = 3
    SERIAL = 4


@dataclasses.dataclass
class CommParams:
    """Parameter class."""

    # generic
    comm_name: str | None = None
    comm_type: CommType | None = None
    reconnect_delay: float | None = None
    reconnect_delay_max: float = 0.0
    timeout_connect: float = 0.0
    host: str = "localhost" # On some machines this will now be ::1
    port: int = 0
    source_address: tuple[str, int] | None = None
    handle_local_echo: bool = False

    # tls
    sslctx: ssl.SSLContext | None = None

    # serial
    baudrate: int = -1
    bytesize: int = -1
    parity: str = ''
    stopbits: int = -1

    @classmethod
    def generate_ssl(
        cls,
        is_server: bool,
        certfile: str | None = None,
        keyfile: str | None = None,
        password: str | None = None,
        sslctx: ssl.SSLContext | None = None,
    ) -> ssl.SSLContext:
        """Generate sslctx from cert/key/password.

        MODBUS/TCP Security Protocol Specification demands TLSv2 at least
        """
        if sslctx:
            return sslctx
        new_sslctx = ssl.SSLContext(
            ssl.PROTOCOL_TLS_SERVER if is_server else ssl.PROTOCOL_TLS_CLIENT
        )
        new_sslctx.check_hostname = False
        new_sslctx.verify_mode = ssl.CERT_NONE
        new_sslctx.minimum_version = ssl.TLSVersion.TLSv1_2
        new_sslctx.maximum_version = ssl.TLSVersion.TLSv1_3
        if certfile:
            new_sslctx.load_cert_chain(
                certfile=certfile, keyfile=keyfile, password=password
            )
        return new_sslctx

    def copy(self) -> CommParams:
        """Create a copy."""
        return dataclasses.replace(self)


class ModbusProtocol(asyncio.BaseProtocol):
    """Protocol layer including transport."""

    def __init__(
        self,
        params: CommParams,
        is_server: bool,
        is_sync: bool = False
    ) -> None:
        """Initialize a transport instance.

        :param params: parameter dataclass
        :param is_server: true if object act as a server (listen/connect)
        :param is_sync: true if used with sync client
        """
        self.comm_params = params.copy()
        self.is_server = is_server
        self.is_closing = False

        self.transport: asyncio.BaseTransport = None  # type: ignore[assignment]
        self.recv_buffer: bytes = b""
        self.call_create: Callable[[], Coroutine[Any, Any, Any]] = None  # type: ignore[assignment]
        self.reconnect_task: asyncio.Task | None = None
        self.listener: ModbusProtocol | None = None
        self.is_listener: bool = False
        self.active_connections: dict[str, ModbusProtocol] = {}
        self.unique_id: str = str(id(self))
        self.reconnect_delay_current = 0.0
        self.sent_buffer: bytes = b""
        self.loop: asyncio.AbstractEventLoop
        if is_sync:
            return
        self.loop = asyncio.get_running_loop()
        if self.is_server:
            if self.comm_params.source_address is not None:
                host = self.comm_params.source_address[0]
                port = int(self.comm_params.source_address[1])
            else:
                # This behaviour isn't quite right.
                # It listens on any IPv4 address rather than the more natural default of any address (v6 or v4).
                host = "0.0.0.0" # Any IPv4 host
                port = 502 # Server will listen on standard modbus port
        else:
            host = self.comm_params.host
            port = int(self.comm_params.port)
        if self.comm_params.comm_type == CommType.SERIAL and NULLMODEM_HOST in host:
            host, port = NULLMODEM_HOST, int(host[9:].split(":")[1])
        if host == NULLMODEM_HOST:
            self.call_create = partial(self.create_nullmodem, port)
            return
        if (
            self.comm_params.comm_type == CommType.SERIAL
            and self.is_server
            and host.startswith("socket")
        ):
            # format is "socket://<host>:port"
            self.comm_params.comm_type = CommType.TCP
            parts = host.split(":")
            host, port = parts[1][2:], int(parts[2])
        self.init_setup_connect_listen(host, port)

    def init_setup_connect_listen(self, host: str, port: int) -> None:
        """Handle connect/listen handler."""
        if self.comm_params.comm_type == CommType.SERIAL:
            self.call_create = partial(create_serial_connection,
                self.loop,
                self.handle_new_connection,
                host,
                baudrate=self.comm_params.baudrate,
                bytesize=self.comm_params.bytesize,
                parity=self.comm_params.parity,
                stopbits=self.comm_params.stopbits,
                timeout=self.comm_params.timeout_connect,
            )
            return
        if self.comm_params.comm_type == CommType.UDP:
            if self.is_server:
                self.call_create = partial(self.loop.create_datagram_endpoint,
                    self.handle_new_connection,
                    local_addr=(host, port),
                )
            else:
                self.call_create = partial(self.loop.create_datagram_endpoint,
                    self.handle_new_connection,
                    remote_addr=(host, port),
                )
            return
        # TLS and TCP
        if self.is_server:
            self.call_create = partial(self.loop.create_server,
                self.handle_new_connection,
                host,
                port,
                ssl=self.comm_params.sslctx,
                reuse_address=True,
                start_serving=True,
            )
        else:
            self.call_create = partial(self.loop.create_connection,
                self.handle_new_connection,
                host,
                port,
                local_addr=self.comm_params.source_address,
                ssl=self.comm_params.sslctx,
            )

    async def connect(self) -> bool:
        """Handle generic connect and call on to specific transport connect."""
        Log.debug("Connecting {}", self.comm_params.comm_name)
        self.is_closing = False
        try:
            self.transport, _protocol = await asyncio.wait_for(
                self.call_create(),
                timeout=self.comm_params.timeout_connect,
            )
        except (asyncio.TimeoutError, OSError) as exc:  # pylint: disable=overlapping-except
            Log.warning("Failed to connect {}", exc)
            return False
        return bool(self.transport)

    async def listen(self) -> bool:
        """Handle generic listen and call on to specific transport listen."""
        Log.debug("Awaiting connections {}", self.comm_params.comm_name)
        self.is_closing = False
        self.is_listener = True
        try:
            self.transport = await self.call_create()
            if isinstance(self.transport, tuple):
                self.transport = self.transport[0]
        except OSError as exc:
            Log.warning("Failed to start server {}", exc)
            self.__close()
            return False
        return True

    # --------------------------------------- #
    # ModbusProtocol asyncio standard methods #
    # --------------------------------------- #
    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        """Call from asyncio, when a connection is made.

        :param transport: socket etc. representing the connection.
        """
        Log.debug("Connected to {}", self.comm_params.comm_name)
        self.transport = transport
        self.reset_delay()
        self.callback_connected()

    def connection_lost(self, reason: Exception | None) -> None:
        """Call from asyncio, when the connection is lost or closed.

        :param reason: None or an exception object
        """
        if not self.transport or self.is_closing:
            return
        Log.debug("Connection lost {} due to {}", self.comm_params.comm_name, reason)
        self.__close()
        if self.is_listener:
            self.reconnect_task = asyncio.create_task(self.do_relisten())
            self.reconnect_task.set_name("transport relisten")
        elif not self.listener and self.comm_params.reconnect_delay:
            self.reconnect_task = asyncio.create_task(self.do_reconnect())
            self.reconnect_task.set_name("transport reconnect")
        self.callback_disconnected(reason)

    def data_received(self, data: bytes) -> None:
        """Call when some data is received.

        :param data: non-empty bytes object with incoming data.
        """
        self.datagram_received(data, None)

    def datagram_received(self, data: bytes, addr: tuple | None) -> None:
        """Receive datagram (UDP connections)."""
        if self.comm_params.handle_local_echo and self.sent_buffer:
            if data.startswith(self.sent_buffer):
                Log.debug(
                    "recv skipping (local_echo): {} addr={}",
                    self.sent_buffer,
                    ":hex",
                    addr,
                )
                data = data[len(self.sent_buffer) :]
                self.sent_buffer = b""
            elif self.sent_buffer.startswith(data):
                Log.debug(
                    "recv skipping (partial local_echo): {} addr={}", data, ":hex", addr
                )
                self.sent_buffer = self.sent_buffer[len(data) :]
                return
            else:
                Log.debug("did not receive local echo: {} addr={}", data, ":hex", addr)
                self.sent_buffer = b""
            if not data:
                return
        Log.debug(
            "recv: {} old_data: {} addr={}",
            data,
            ":hex",
            self.recv_buffer,
            ":hex",
            addr,
        )
        self.recv_buffer += data
        cut = self.callback_data(self.recv_buffer, addr=addr)
        self.recv_buffer = self.recv_buffer[cut:]
        if self.recv_buffer:
            Log.debug(
                "recv, unused data waiting for next packet: {}",
                self.recv_buffer,
                ":hex",
            )

    def eof_received(self) -> None:
        """Accept other end terminates connection."""
        Log.debug("-> transport: received eof")

    def error_received(self, exc):
        """Get error detected in UDP."""
        Log.debug("-> error_received {}", exc)

    # --------- #
    # callbacks #
    # --------- #
    @abstractmethod
    def callback_new_connection(self) -> ModbusProtocol:
        """Call when listener receive new connection request."""

    @abstractmethod
    def callback_connected(self) -> None:
        """Call when connection is succcesfull."""

    @abstractmethod
    def callback_disconnected(self, exc: Exception | None) -> None:
        """Call when connection is lost."""

    @abstractmethod
    def callback_data(self, data: bytes, addr: tuple | None = None) -> int:
        """Handle received data."""

    # ----------------------------------- #
    # Helper methods for external classes #
    # ----------------------------------- #
    def send(self, data: bytes, addr: tuple | None = None) -> None:
        """Send modbus message.

        :param data: non-empty bytes object with data to send.
        :param addr: optional addr, only used for UDP server.
        """
        if not self.transport:
            Log.error("Cancel send, because not connected!")
            return
        Log.debug("send: {}", data, ":hex")
        self.recv_buffer = b""
        if self.comm_params.handle_local_echo:
            self.sent_buffer += data
        if self.comm_params.comm_type == CommType.UDP:
            if addr:
                self.transport.sendto(data, addr=addr)  # type: ignore[attr-defined]
            else:
                self.transport.sendto(data)  # type: ignore[attr-defined]
        else:
            self.transport.write(data)  # type: ignore[attr-defined]

    def __close(self, reconnect: bool = False) -> None:
        """Close connection (internal).

        :param reconnect: (default false), try to reconnect
        """
        if self.transport:
            self.transport.close()
            self.transport = None  # type: ignore[assignment]
        self.recv_buffer = b""
        if self.is_listener:
            for _key, value in self.active_connections.items():
                value.listener = None
                value.callback_disconnected(None)
                value.close()
            self.active_connections = {}
            return
        if not reconnect and self.reconnect_task:
            self.reconnect_task.cancel()
            self.reconnect_task = None
            self.reconnect_delay_current = 0.0
        if self.listener:
            self.listener.active_connections.pop(self.unique_id)

    def close(self, reconnect: bool = False) -> None:
        """Close connection (external).

        :param reconnect: (default false), try to reconnect
        """
        if self.is_closing:
            return
        self.is_closing = True
        self.__close(reconnect=reconnect)

    def reset_delay(self) -> None:
        """Reset wait time before next reconnect to minimal period."""
        self.reconnect_delay_current = self.comm_params.reconnect_delay or 0.0

    def is_active(self) -> bool:
        """Return true if connected/listening."""
        return bool(self.transport)

    # ---------------- #
    # Internal methods #
    # ---------------- #
    async def create_nullmodem(
        self, port
    ) -> tuple[asyncio.Transport, asyncio.BaseProtocol]:
        """Bypass create_ and use null modem."""
        if self.is_server:
            # Listener object
            self.transport = NullModem.set_listener(port, self)
            return self.transport, self

        # connect object
        return NullModem.set_connection(port, self)

    def handle_new_connection(self) -> ModbusProtocol:
        """Handle incoming connect."""
        if not self.is_server:
            # Clients reuse the same object.
            return self

        new_protocol = self.callback_new_connection()
        self.active_connections[new_protocol.unique_id] = new_protocol
        new_protocol.listener = self
        return new_protocol

    async def do_relisten(self) -> None:
        """Handle reconnect as a task."""
        try:
            Log.debug("Wait 1s before reopening listener.")
            await asyncio.sleep(1)
            await self.listen()
        except asyncio.CancelledError:
            pass
        self.reconnect_task = None

    async def do_reconnect(self) -> None:
        """Handle reconnect as a task."""
        try:
            self.reconnect_delay_current = self.comm_params.reconnect_delay or 0.0
            while True:
                Log.debug(
                    "Wait {} {} ms before reconnecting.",
                    self.comm_params.comm_name,
                    self.reconnect_delay_current * 1000,
                )
                await asyncio.sleep(self.reconnect_delay_current)
                if await self.connect():
                    break
                self.reconnect_delay_current = min(
                    2 * self.reconnect_delay_current,
                    self.comm_params.reconnect_delay_max,
                )
        except asyncio.CancelledError:
            pass
        self.reconnect_task = None

    # ----------------- #
    # The magic methods #
    # ----------------- #
    async def __aenter__(self) -> ModbusProtocol:
        """Implement the client with async enter block."""
        return self

    async def __aexit__(self, _class, _value, _traceback) -> None:
        """Implement the client with async exit block."""
        self.close()

    def __str__(self) -> str:
        """Build a string representation of the connection."""
        return f"{self.__class__.__name__}({self.comm_params.comm_name})"


class NullModem(asyncio.DatagramTransport, asyncio.Transport):
    """ModbusProtocol layer.

    Contains methods to act as a null modem between 2 objects.
    (Allowing tests to be shortcut without actual network calls)
    """

    listeners: dict[int, ModbusProtocol] = {}
    connections: dict[NullModem, int] = {}

    def __init__(self, protocol: ModbusProtocol, listen: int | None = None) -> None:
        """Create half part of null modem."""
        asyncio.DatagramTransport.__init__(self)
        asyncio.Transport.__init__(self)
        self.protocol: ModbusProtocol = protocol
        self.other_modem: NullModem = None  # type: ignore[assignment]
        self.listen = listen
        self.manipulator: Callable[[bytes], list[bytes]] | None = None
        self._is_closing = False

    # -------------------------- #
    # external nullmodem methods #
    # -------------------------- #
    @classmethod
    def set_listener(cls, port: int, parent: ModbusProtocol) -> NullModem:
        """Register listener."""
        if port in cls.listeners:
            raise AssertionError(f"Port {port} already listening !")
        cls.listeners[port] = parent
        return NullModem(parent, listen=port)

    @classmethod
    def set_connection(
        cls, port: int, parent: ModbusProtocol
    ) -> tuple[NullModem, ModbusProtocol]:
        """Connect to listener."""
        if port not in cls.listeners:
            raise asyncio.TimeoutError(f"Port {port} not being listened on !")

        client_protocol = parent.handle_new_connection()
        server_protocol = NullModem.listeners[port].handle_new_connection()
        client_transport = NullModem(client_protocol)
        server_transport = NullModem(server_protocol)
        cls.connections[client_transport] = port
        cls.connections[server_transport] = -port
        client_transport.other_modem = server_transport
        server_transport.other_modem = client_transport
        client_protocol.connection_made(client_transport)
        server_protocol.connection_made(server_transport)
        return client_transport, client_protocol

    def set_manipulator(self, function: Callable[[bytes], list[bytes]]) -> None:
        """Register a manipulator."""
        self.manipulator = function

    @classmethod
    def is_dirty(cls):
        """Check if everything is closed."""
        dirty = False
        if cls.connections:
            Log.error(
                "NullModem_FATAL missing close on port {} connect()",
                [str(key) for key in cls.connections.values()],
            )
            dirty = True
        if cls.listeners:
            Log.error(
                "NullModem_FATAL missing close on port {} listen()",
                [str(value) for value in cls.listeners],
            )
            dirty = True
        return dirty

    # ---------------- #
    # external methods #
    # ---------------- #

    def close(self) -> None:
        """Close null modem."""
        if self._is_closing:
            return
        self._is_closing = True
        if self.listen:
            del self.listeners[self.listen]
            return
        if self.connections:
            with suppress(KeyError):
                del self.connections[self]
        if self.other_modem:
            self.other_modem.other_modem = None  # type: ignore[assignment]
            self.other_modem.close()
            self.other_modem = None  # type: ignore[assignment]
        if self.protocol:
            self.protocol.connection_lost(None)

    def sendto(self, data: bytes, _addr: Any = None) -> None:
        """Send datagrame."""
        self.write(data)

    def write(self, data: bytes) -> None:
        """Send data."""
        if not self.manipulator:
            if self.other_modem:
                self.other_modem.protocol.data_received(data)
            return
        data_manipulated = self.manipulator(data)
        for part in data_manipulated:
            self.other_modem.protocol.data_received(part)

    # ------------- #
    # Dummy methods #
    # ------------- #
    def abort(self) -> None:
        """Alias for closing the connection."""
        self.close()

    def can_write_eof(self) -> bool:
        """Allow to write eof."""
        return False

    def get_write_buffer_size(self) -> int:
        """Set write limit."""
        return 1024

    def get_write_buffer_limits(self) -> tuple[int, int]:
        """Set flush limits."""
        return (1, 1024)

    def set_write_buffer_limits(
        self, high: int | None = None, low: int | None = None
    ) -> None:
        """Set flush limits."""

    def write_eof(self) -> None:
        """Write eof."""

    def get_protocol(self) -> ModbusProtocol | asyncio.BaseProtocol:
        """Return current protocol."""
        return self.protocol

    def set_protocol(self, protocol: asyncio.BaseProtocol) -> None:
        """Set current protocol."""

    def is_closing(self) -> bool:
        """Return true if closing."""
        return self._is_closing

    def is_reading(self) -> bool:
        """Return true if read is active."""
        return True

    def pause_reading(self):
        """Pause receiver."""

    def resume_reading(self):
        """Resume receiver."""



"""Modbus Device Controller.

These are the device management handlers.  They should be
maintained in the server context and the various methods
should be inserted in the correct locations.
"""
from __future__ import annotations


__all__ = [
    "DeviceInformationFactory",
    "ModbusDeviceIdentification",
    "ModbusPlusStatistics",
]

import struct

# pylint: disable=missing-type-doc
from collections import OrderedDict

from pymodbus.constants import INTERNAL_ERROR, DeviceInformation
from pymodbus.events import ModbusEvent
from pymodbus.utilities import dict_property


# ---------------------------------------------------------------------------#
#  Modbus Plus Statistics
# ---------------------------------------------------------------------------#
class ModbusPlusStatistics:
    """This is used to maintain the current modbus plus statistics count.

    As of right now this is simply a stub to complete the modbus implementation.
    For more information, see the modbus implementation guide page 87.
    """

    __data = OrderedDict(
        {
            "node_type_id": [0x00] * 2,  # 00
            "software_version_number": [0x00] * 2,  # 01
            "network_address": [0x00] * 2,  # 02
            "mac_state_variable": [0x00] * 2,  # 03
            "peer_status_code": [0x00] * 2,  # 04
            "token_pass_counter": [0x00] * 2,  # 05
            "token_rotation_time": [0x00] * 2,  # 06
            "program_master_token_failed": [0x00],  # 07 hi
            "data_master_token_failed": [0x00],  # 07 lo
            "program_master_token_owner": [0x00],  # 08 hi
            "data_master_token_owner": [0x00],  # 08 lo
            "program_slave_token_owner": [0x00],  # 09 hi
            "data_slave_token_owner": [0x00],  # 09 lo
            "data_slave_command_transfer": [0x00],  # 10 hi
            "__unused_10_lowbit": [0x00],  # 10 lo
            "program_slave_command_transfer": [0x00],  # 11 hi
            "program_master_rsp_transfer": [0x00],  # 11 lo
            "program_slave_auto_logout": [0x00],  # 12 hi
            "program_master_connect_status": [0x00],  # 12 lo
            "receive_buffer_dma_overrun": [0x00],  # 13 hi
            "pretransmit_deferral_error": [0x00],  # 13 lo
            "frame_size_error": [0x00],  # 14 hi
            "repeated_command_received": [0x00],  # 14 lo
            "receiver_alignment_error": [0x00],  # 15 hi
            "receiver_collision_abort_error": [0x00],  # 15 lo
            "bad_packet_length_error": [0x00],  # 16 hi
            "receiver_crc_error": [0x00],  # 16 lo
            "transmit_buffer_dma_underrun": [0x00],  # 17 hi
            "bad_link_address_error": [0x00],  # 17 lo
            "bad_mac_function_code_error": [0x00],  # 18 hi
            "internal_packet_length_error": [0x00],  # 18 lo
            "communication_failed_error": [0x00],  # 19 hi
            "communication_retries": [0x00],  # 19 lo
            "no_response_error": [0x00],  # 20 hi
            "good_receive_packet": [0x00],  # 20 lo
            "unexpected_path_error": [0x00],  # 21 hi
            "exception_response_error": [0x00],  # 21 lo
            "forgotten_transaction_error": [0x00],  # 22 hi
            "unexpected_response_error": [0x00],  # 22 lo
            "active_station_bit_map": [0x00] * 8,  # 23-26
            "token_station_bit_map": [0x00] * 8,  # 27-30
            "global_data_bit_map": [0x00] * 8,  # 31-34
            "receive_buffer_use_bit_map": [0x00] * 8,  # 35-37
            "data_master_output_path": [0x00] * 8,  # 38-41
            "data_slave_input_path": [0x00] * 8,  # 42-45
            "program_master_outptu_path": [0x00] * 8,  # 46-49
            "program_slave_input_path": [0x00] * 8,  # 50-53
        }
    )

    def __init__(self):
        """Initialize the modbus plus statistics with the default information."""
        self.reset()

    def __iter__(self):
        """Iterate over the statistics.

        :returns: An iterator of the modbus plus statistics
        """
        return iter(self.__data.items())

    def reset(self):
        """Clear all of the modbus plus statistics."""
        for key in self.__data:
            self.__data[key] = [0x00] * len(self.__data[key])

    def summary(self):
        """Return a summary of the modbus plus statistics.

        :returns: 54 16-bit words representing the status
        """
        return iter(self.__data.values())

    def encode(self):
        """Return a summary of the modbus plus statistics.

        :returns: 54 16-bit words representing the status
        """
        total, values = [], sum(self.__data.values(), [])  # noqa: RUF017
        for i in range(0, len(values), 2):
            total.append((values[i] << 8) | values[i + 1])
        return total


# ---------------------------------------------------------------------------#
#  Device Information Control
# ---------------------------------------------------------------------------#
class ModbusDeviceIdentification:
    """This is used to supply the device identification.

    For the readDeviceIdentification function

    For more information read section 6.21 of the modbus
    application protocol.
    """

    __data = {
        0x00: "",  # VendorName
        0x01: "",  # ProductCode
        0x02: "",  # MajorMinorRevision
        0x03: "",  # VendorUrl
        0x04: "",  # ProductName
        0x05: "",  # ModelName
        0x06: "",  # UserApplicationName
        0x07: "",  # reserved
        0x08: "",  # reserved
        # 0x80 -> 0xFF are privatek
    }

    __names = [
        "VendorName",
        "ProductCode",
        "MajorMinorRevision",
        "VendorUrl",
        "ProductName",
        "ModelName",
        "UserApplicationName",
    ]

    def __init__(self, info=None, info_name=None):
        """Initialize the datastore with the elements you need.

        (note acceptable range is [0x00-0x06,0x80-0xFF] inclusive)

        :param info: A dictionary of {int:string} of values
        :param set: A dictionary of {name:string} of values
        """
        if isinstance(info_name, dict):
            for key in info_name:
                inx = self.__names.index(key)
                self.__data[inx] = info_name[key]

        if isinstance(info, dict):
            for key in info:
                if (0x06 >= key >= 0x00) or (0xFF >= key >= 0x80):
                    self.__data[key] = info[key]

    def __iter__(self):
        """Iterate over the device information.

        :returns: An iterator of the device information
        """
        return iter(self.__data.items())

    def summary(self):
        """Return a summary of the main items.

        :returns: An dictionary of the main items
        """
        return dict(zip(self.__names, iter(self.__data.values())))

    def update(self, value):
        """Update the values of this identity.

        using another identify as the value

        :param value: The value to copy values from
        """
        self.__data.update(value)

    def __setitem__(self, key, value):
        """Access the device information.

        :param key: The register to set
        :param value: The new value for referenced register
        """
        if key not in [0x07, 0x08]:
            self.__data[key] = value

    def __getitem__(self, key):
        """Access the device information.

        :param key: The register to read
        """
        return self.__data.setdefault(key, "")

    def __str__(self):
        """Build a representation of the device.

        :returns: A string representation of the device
        """
        return "DeviceIdentity"

    # -------------------------------------------------------------------------#
    #  Properties
    # -------------------------------------------------------------------------#
    # fmt: off
    VendorName = dict_property(lambda s: s.__data, 0)  # pylint: disable=protected-access
    ProductCode = dict_property(lambda s: s.__data, 1)  # pylint: disable=protected-access
    MajorMinorRevision = dict_property(lambda s: s.__data, 2)  # pylint: disable=protected-access
    VendorUrl = dict_property(lambda s: s.__data, 3)  # pylint: disable=protected-access
    ProductName = dict_property(lambda s: s.__data, 4)  # pylint: disable=protected-access
    ModelName = dict_property(lambda s: s.__data, 5)  # pylint: disable=protected-access
    UserApplicationName = dict_property(lambda s: s.__data, 6)  # pylint: disable=protected-access
    # fmt: on


class DeviceInformationFactory:  # pylint: disable=too-few-public-methods
    """This is a helper.

    That really just hides
    some of the complexity of processing the device information
    requests (function code 0x2b 0x0e).
    """

    __lookup = {
        DeviceInformation.BASIC: lambda c, r, i: c.__gets(  # pylint: disable=protected-access
            r, list(range(i, 0x03))
        ),
        DeviceInformation.REGULAR: lambda c, r, i: c.__gets(  # pylint: disable=protected-access
            r,
            list(range(i, 0x07))
            if c.__get(r, i)[i]  # pylint: disable=protected-access
            else list(range(0, 0x07)),
        ),
        DeviceInformation.EXTENDED: lambda c, r, i: c.__gets(  # pylint: disable=protected-access
            r,
            [x for x in range(i, 0x100) if x not in range(0x07, 0x80)]
            if c.__get(r, i)[i]  # pylint: disable=protected-access
            else [x for x in range(0, 0x100) if x not in range(0x07, 0x80)],
        ),
        DeviceInformation.SPECIFIC: lambda c, r, i: c.__get(  # pylint: disable=protected-access
            r, i
        ),
    }

    @classmethod
    def get(cls, control, read_code=DeviceInformation.BASIC, object_id=0x00):
        """Get the requested device data from the system.

        :param control: The control block to pull data from
        :param read_code: The read code to process
        :param object_id: The specific object_id to read
        :returns: The requested data (id, length, value)
        """
        identity = control.Identity
        return cls.__lookup[read_code](cls, identity, object_id)

    @classmethod
    def __get(cls, identity, object_id):  # pylint: disable=unused-private-member
        """Read a single object_id from the device information.

        :param identity: The identity block to pull data from
        :param object_id: The specific object id to read
        :returns: The requested data (id, length, value)
        """
        return {object_id: identity[object_id]}

    @classmethod
    def __gets(cls, identity, object_ids):  # pylint: disable=unused-private-member
        """Read multiple object_ids from the device information.

        :param identity: The identity block to pull data from
        :param object_ids: The specific object ids to read
        :returns: The requested data (id, length, value)
        """
        return {oid: identity[oid] for oid in object_ids if identity[oid]}

    def __init__(self):
        """Prohibit objects."""
        raise RuntimeError(INTERNAL_ERROR)


# ---------------------------------------------------------------------------#
#  Counters Handler
# ---------------------------------------------------------------------------#
class ModbusCountersHandler:
    """This is a helper class to simplify the properties for the counters.

    0x0B  1  Return Bus Message Count

             Quantity of messages that the remote
             device has detected on the communications system since its
             last restart, clear counters operation, or power-up.  Messages
             with bad CRC are not taken into account.

    0x0C  2  Return Bus Communication Error Count

             Quantity of CRC errors encountered by the remote device since its
             last restart, clear counters operation, or power-up.  In case of
             an error detected on the character level, (overrun, parity error),
             or in case of a message length < 3 bytes, the receiving device is
             not able to calculate the CRC. In such cases, this counter is
             also incremented.

    0x0D  3  Return Slave Exception Error Count

             Quantity of MODBUS exception error detected by the remote device
             since its last restart, clear counters operation, or power-up.
             Exception errors are described and listed in "MODBUS Application
             Protocol Specification" document.

    0xOE  4  Return Slave Message Count

             Quantity of messages addressed to the remote device that the remote
             device has processed since its last restart, clear counters operation,
             or power-up.

    0x0F  5  Return Slave No Response Count

             Quantity of messages received by the remote device for which it
             returned no response (neither a normal response nor an exception
             response), since its last restart, clear counters operation, or
             power-up.

    0x10  6  Return Slave NAK Count

             Quantity of messages addressed to the remote device for which it
             returned a Negative ACKNOWLEDGE (NAK) exception response, since
             its last restart, clear counters operation, or power-up. Exception
             responses are described and listed in "MODBUS Application Protocol
             Specification" document.

    0x11  7  Return Slave Busy Count

             Quantity of messages addressed to the remote device for which it
             returned a Slave Device Busy exception response, since its last
             restart, clear counters operation, or power-up. Exception
             responses are described and listed in "MODBUS Application
             Protocol Specification" document.

    0x12  8  Return Bus Character Overrun Count

             Quantity of messages addressed to the remote device that it could
             not handle due to a character overrun condition, since its last
             restart, clear counters operation, or power-up. A character
             overrun is caused by data characters arriving at the port faster
             than they can.

    .. note:: I threw the event counter in here for convenience
    """

    __data = dict.fromkeys(range(9), 0x00)
    __names = [
        "BusMessage",
        "BusCommunicationError",
        "SlaveExceptionError",
        "SlaveMessage",
        "SlaveNoResponse",
        "SlaveNAK",
        "SLAVE_BUSY",
        "BusCharacterOverrun",
    ]

    def __iter__(self):
        """Iterate over the device counters.

        :returns: An iterator of the device counters
        """
        return zip(self.__names, iter(self.__data.values()))

    def update(self, values):
        """Update the values of this identity.

        using another identify as the value

        :param values: The value to copy values from
        """
        for k, v_item in iter(values.items()):
            v_item += self.__getattribute__(  # pylint: disable=unnecessary-dunder-call
                k
            )
            self.__setattr__(k, v_item)  # pylint: disable=unnecessary-dunder-call

    def reset(self):
        """Clear all of the system counters."""
        self.__data = dict.fromkeys(range(9), 0x00)

    def summary(self):
        """Return a summary of the counters current status.

        :returns: A byte with each bit representing each counter
        """
        count, result = 0x01, 0x00
        for i in iter(self.__data.values()):
            if i != 0x00:  # pylint: disable=compare-to-zero
                result |= count
            count <<= 1
        return result

    # -------------------------------------------------------------------------#
    #  Properties
    # -------------------------------------------------------------------------#
    # fmt: off
    BusMessage = dict_property(lambda s: s.__data, 0)  # pylint: disable=protected-access
    BusCommunicationError = dict_property(lambda s: s.__data, 1)  # pylint: disable=protected-access
    BusExceptionError = dict_property(lambda s: s.__data, 2)  # pylint: disable=protected-access
    SlaveMessage = dict_property(lambda s: s.__data, 3)  # pylint: disable=protected-access
    SlaveNoResponse = dict_property(lambda s: s.__data, 4)  # pylint: disable=protected-access
    SlaveNAK = dict_property(lambda s: s.__data, 5)  # pylint: disable=protected-access
    SLAVE_BUSY = dict_property(lambda s: s.__data, 6)  # pylint: disable=protected-access
    BusCharacterOverrun = dict_property(lambda s: s.__data, 7)  # pylint: disable=protected-access
    Event = dict_property(lambda s: s.__data, 8)  # pylint: disable=protected-access
    # fmt: on


# ---------------------------------------------------------------------------#
#  Main server control block
# ---------------------------------------------------------------------------#
class ModbusControlBlock:
    """This is a global singleton that controls all system information.

    All activity should be logged here and all diagnostic requests
    should come from here.
    """

    _mode = "ASCII"
    _diagnostic = [False] * 16
    _listen_only = False
    _delimiter = b"\r"
    _counters = ModbusCountersHandler()
    _identity = ModbusDeviceIdentification()
    _plus = ModbusPlusStatistics()
    _events: list[ModbusEvent] = []

    # -------------------------------------------------------------------------#
    #  Magic
    # -------------------------------------------------------------------------#
    def __str__(self):
        """Build a representation of the control block.

        :returns: A string representation of the control block
        """
        return "ModbusControl"

    def __iter__(self):
        """Iterate over the device counters.

        :returns: An iterator of the device counters
        """
        return self._counters.__iter__()

    def __new__(cls):
        """Create a new instance."""
        if "_inst" not in vars(cls):
            cls._inst = object.__new__(cls)
        return cls._inst

    # -------------------------------------------------------------------------#
    #  Events
    # -------------------------------------------------------------------------#
    def addEvent(self, event: ModbusEvent):
        """Add a new event to the event log.

        :param event: A new event to add to the log
        """
        self._events.insert(0, event)
        self._events = self._events[0:64]  # chomp to 64 entries
        self.Counter.Event += 1

    def getEvents(self):
        """Return an encoded collection of the event log.

        :returns: The encoded events packet
        """
        events = [event.encode() for event in self._events]
        return b"".join(events)

    def clearEvents(self):
        """Clear the current list of events."""
        self._events = []

    # -------------------------------------------------------------------------#
    #  Other Properties
    # -------------------------------------------------------------------------#
    Identity = property(lambda s: s._identity)
    Counter = property(lambda s: s._counters)
    Events = property(lambda s: s._events)
    Plus = property(lambda s: s._plus)

    def reset(self):
        """Clear all of the system counters and the diagnostic register."""
        self._events = []
        self._counters.reset()
        self._diagnostic = [False] * 16

    # -------------------------------------------------------------------------#
    #  Listen Properties
    # -------------------------------------------------------------------------#
    def _setListenOnly(self, value):
        """Toggle the listen only status.

        :param value: The value to set the listen status to
        """
        self._listen_only = bool(value)

    ListenOnly = property(lambda s: s._listen_only, _setListenOnly)

    # -------------------------------------------------------------------------#
    #  Mode Properties
    # -------------------------------------------------------------------------#
    def _setMode(self, mode):
        """Toggle the current serial mode.

        :param mode: The data transfer method in (RTU, ASCII)
        """
        if mode in {"ASCII", "RTU"}:
            self._mode = mode

    Mode = property(lambda s: s._mode, _setMode)

    # -------------------------------------------------------------------------#
    #  Delimiter Properties
    # -------------------------------------------------------------------------#
    def _setDelimiter(self, char):
        """Change the serial delimiter character.

        :param char: The new serial delimiter character
        """
        if isinstance(char, str):
            self._delimiter = char.encode()
        if isinstance(char, bytes):
            self._delimiter = char
        elif isinstance(char, int):
            self._delimiter = struct.pack(">B", char)

    Delimiter = property(lambda s: s._delimiter, _setDelimiter)

    # -------------------------------------------------------------------------#
    #  Diagnostic Properties
    # -------------------------------------------------------------------------#
    def setDiagnostic(self, mapping):
        """Set the value in the diagnostic register.

        :param mapping: Dictionary of key:value pairs to set
        """
        for entry in iter(mapping.items()):
            if entry[0] >= 0 and entry[0] < len(self._diagnostic):
                self._diagnostic[entry[0]] = bool(entry[1])

    def getDiagnostic(self, bit):
        """Get the value in the diagnostic register.

        :param bit: The bit to get
        :returns: The current value of the requested bit
        """
        try:
            if bit and 0 <= bit < len(self._diagnostic):
                return self._diagnostic[bit]
        except Exception:  # pylint: disable=broad-except
            return None
        return None

    def getDiagnosticRegister(self):
        """Get the entire diagnostic register.

        :returns: The diagnostic register collection
        """
        return self._diagnostic



"""Bit Reading Request/Response messages."""

import struct
from typing import cast

from pymodbus.constants import ModbusStatus
from pymodbus.datastore import ModbusSlaveContext
from pymodbus.pdu.pdu import (
    ExceptionResponse,
    ModbusPDU,
    pack_bitstring,
    unpack_bitstring,
)


class ReadCoilsRequest(ModbusPDU):
    """ReadCoilsRequest."""

    rtu_frame_size = 8
    function_code = 1

    def encode(self) -> bytes:
        """Encode a request pdu."""
        self.verifyCount(2000)
        self.verifyAddress()
        return struct.pack(">HH", self.address, self.count)

    def decode(self, data: bytes) -> None:
        """Decode a request pdu."""
        self.address, self.count = struct.unpack(">HH", data)

    def get_response_pdu_size(self) -> int:
        """Get response pdu size.

        Func_code (1 byte) + Byte Count(1 byte) + Quantity of Coils (n Bytes)/8,
        if the remainder is different of 0 then N = N+1
        """
        return 1 + 1 + (self.count + 7) // 8

    async def update_datastore(self, context: ModbusSlaveContext) -> ModbusPDU:
        """Run request against a datastore."""
        values = await context.async_getValues(
            self.function_code, self.address, self.count
        )
        if isinstance(values, int):
            return ExceptionResponse(self.function_code, values)
        response_class = (ReadCoilsResponse if self.function_code == 1 else ReadDiscreteInputsResponse)
        return response_class(dev_id=self.dev_id, transaction_id=self.transaction_id, bits=cast(list[bool], values))


class ReadDiscreteInputsRequest(ReadCoilsRequest):
    """ReadDiscreteInputsRequest."""

    function_code = 2


class ReadCoilsResponse(ModbusPDU):
    """ReadCoilsResponse."""

    function_code = 1
    rtu_byte_count_pos = 2

    def encode(self) -> bytes:
        """Encode response pdu."""
        result = pack_bitstring(self.bits)
        return struct.pack(">B", len(result)) + result

    def decode(self, data: bytes) -> None:
        """Decode response pdu."""
        self.bits = unpack_bitstring(data[1:])


class ReadDiscreteInputsResponse(ReadCoilsResponse):
    """ReadDiscreteInputsResponse."""

    function_code = 2


class WriteSingleCoilResponse(ModbusPDU):
    """WriteSingleCoilResponse."""

    function_code = 5
    rtu_frame_size = 8

    def encode(self) -> bytes:
        """Encode write coil request."""
        self.verifyAddress()
        val = ModbusStatus.ON if self.bits[0] else ModbusStatus.OFF
        return struct.pack(">HH", self.address, val)

    def decode(self, data: bytes) -> None:
        """Decode a write coil request."""
        self.address, value = struct.unpack(">HH", data)
        self.bits = [bool(value)]


class WriteSingleCoilRequest(WriteSingleCoilResponse):
    """WriteSingleCoilRequest."""

    async def update_datastore(self, context: ModbusSlaveContext) -> ModbusPDU:
        """Run a request against a datastore."""
        if (rc := await context.async_setValues(self.function_code, self.address, self.bits)):
            return ExceptionResponse(self.function_code, rc)
        values = await context.async_getValues(self.function_code, self.address, 1)
        if isinstance(values, int):
            return ExceptionResponse(self.function_code, values)

        return WriteSingleCoilResponse(address=self.address, bits=cast(list[bool], values), dev_id=self.dev_id, transaction_id=self.transaction_id)

    def get_response_pdu_size(self) -> int:
        """Get response pdu size.

        Func_code (1 byte) + Output Address (2 byte) + Output Value  (2 Bytes)
        """
        return 1 + 2 + 2


class WriteMultipleCoilsRequest(ModbusPDU):
    """WriteMultipleCoilsRequest."""

    function_code = 15
    rtu_byte_count_pos = 6

    def encode(self) -> bytes:
        """Encode write coils request."""
        self.count = len(self.bits)
        self.verifyAddress()
        self.verifyCount(2000)
        byte_count = (self.count + 7) // 8
        return struct.pack(">HHB", self.address, self.count, byte_count) + pack_bitstring(self.bits)

    def decode(self, data: bytes) -> None:
        """Decode a write coils request."""
        self.address, count, _byte_count = struct.unpack(">HHB", data[0:5])
        self.bits = unpack_bitstring(data[5:])[:count]

    async def update_datastore(self, context: ModbusSlaveContext) -> ModbusPDU:
        """Run a request against a datastore."""
        count = len(self.bits)
        rc = await context.async_setValues(
            self.function_code, self.address, self.bits
        )
        if rc:
            return ExceptionResponse(self.function_code, rc)

        return WriteMultipleCoilsResponse(address=self.address, count=count, dev_id=self.dev_id, transaction_id=self.transaction_id)

    def get_response_pdu_size(self) -> int:
        """Get response pdu size.

        Func_code (1 byte) + Output Address (2 byte) + Quantity of Outputs  (2 Bytes)
        :return:
        """
        return 1 + 2 + 2


class WriteMultipleCoilsResponse(ModbusPDU):
    """WriteMultipleCoilsResponse."""

    function_code = 15
    rtu_frame_size = 8

    def encode(self) -> bytes:
        """Encode write coils response."""
        return struct.pack(">HH", self.address, self.count)

    def decode(self, data: bytes) -> None:
        """Decode a write coils response."""
        self.address, self.count = struct.unpack(">HH", data)



"""Implementation of a Threaded Modbus Server."""
from __future__ import annotations

import asyncio
import traceback

from pymodbus.exceptions import ModbusIOException, NoSuchSlaveException
from pymodbus.logging import Log
from pymodbus.pdu.pdu import ExceptionResponse
from pymodbus.transaction import TransactionManager
from pymodbus.transport import CommParams, ModbusProtocol


class ServerRequestHandler(TransactionManager):
    """Handle client connection."""

    def __init__(self, owner, trace_packet, trace_pdu, trace_connect):
        """Initialize."""
        params = CommParams(
            comm_name="server",
            comm_type=owner.comm_params.comm_type,
            reconnect_delay=0.0,
            reconnect_delay_max=0.0,
            timeout_connect=0.0,
            host=owner.comm_params.source_address[0],
            port=owner.comm_params.source_address[1],
            handle_local_echo=owner.comm_params.handle_local_echo,
        )
        self.server = owner
        self.framer = self.server.framer(self.server.decoder)
        self.running = False
        super().__init__(
            params,
            self.framer,
            0,
            True,
            trace_packet,
            trace_pdu,
            trace_connect,
        )

    def callback_new_connection(self) -> ModbusProtocol:
        """Call when listener receive new connection request."""
        raise RuntimeError("callback_new_connection should never be called")

    def callback_connected(self) -> None:
        """Call when connection is succcesfull."""
        super().callback_connected()
        slaves = self.server.context.slaves()
        if self.server.broadcast_enable:
            if 0 not in slaves:
                slaves.append(0)

    def callback_disconnected(self, call_exc: Exception | None) -> None:
        """Call when connection is lost."""
        super().callback_disconnected(call_exc)
        try:
            if call_exc is None:
                Log.debug(
                    "Handler for stream [{}] has been canceled", self.comm_params.comm_name
                )
            else:
                Log.debug(
                    "Client Disconnection {} due to {}",
                    self.comm_params.comm_name,
                    call_exc,
                )
            self.running = False
        except Exception as exc:  # pylint: disable=broad-except
            Log.error(
                "Datastore unable to fulfill request: {}; {}",
                exc,
                traceback.format_exc(),
            )

    def callback_data(self, data: bytes, addr: tuple | None = None) -> int:
        """Handle received data."""
        try:
            used_len = super().callback_data(data, addr)
        except ModbusIOException:
            response = ExceptionResponse(
                40,
                exception_code=ExceptionResponse.ILLEGAL_FUNCTION
            )
            self.server_send(response, 0)
            return(len(data))
        if self.last_pdu:
            self.loop.call_soon(self.handle_later)
        return used_len

    def handle_later(self):
        """Change sync (async not allowed in call_soon) to async."""
        asyncio.run_coroutine_threadsafe(self.handle_request(), self.loop)

    async def handle_request(self):
        """Handle request."""
        broadcast = False
        if not self.last_pdu:
            return
        try:
            if self.server.broadcast_enable and not self.last_pdu.dev_id:
                broadcast = True
                # if broadcasting then execute on all slave contexts,
                # note response will be ignored
                for dev_id in self.server.context.slaves():
                    response = await self.last_pdu.update_datastore(self.server.context[dev_id])
            else:
                context = self.server.context[self.last_pdu.dev_id]
                response = await self.last_pdu.update_datastore(context)

        except NoSuchSlaveException:
            Log.error("requested slave does not exist: {}", self.last_pdu.dev_id)
            if self.server.ignore_missing_slaves:
                return  # the client will simply timeout waiting for a response
            response = ExceptionResponse(0x00, ExceptionResponse.GATEWAY_NO_RESPONSE)
        except Exception as exc:  # pylint: disable=broad-except
            Log.error(
                "Datastore unable to fulfill request: {}; {}",
                exc,
                traceback.format_exc(),
            )
            response = ExceptionResponse(0x00, ExceptionResponse.SLAVE_FAILURE)
        # no response when broadcasting
        if not broadcast:
            response.transaction_id = self.last_pdu.transaction_id
            response.dev_id = self.last_pdu.dev_id
            self.server_send(response, self.last_addr)

    def server_send(self, pdu, addr):
        """Send message."""
        if not pdu:
            Log.debug("Skipping sending response!!")
        else:
            self.pdu_send(pdu, addr=addr)



"""Simulator."""

__all__ = [
    "SimAction",
    "SimCore",
    "SimData",
    "SimDataType",
    "SimDevice",
    "SimValueType",
]

from pymodbus.simulator.simcore import SimCore
from pymodbus.simulator.simdata import (
    SimAction,
    SimData,
    SimDataType,
    SimDevice,
    SimValueType,
)



"""Download files with progress indicators.
"""

import email.message
import logging
import mimetypes
import os
from typing import Iterable, Optional, Tuple

from pip._vendor.requests.models import Response

from pip._internal.cli.progress_bars import get_download_progress_renderer
from pip._internal.exceptions import NetworkConnectionError
from pip._internal.models.index import PyPI
from pip._internal.models.link import Link
from pip._internal.network.cache import is_from_cache
from pip._internal.network.session import PipSession
from pip._internal.network.utils import HEADERS, raise_for_status, response_chunks
from pip._internal.utils.misc import format_size, redact_auth_from_url, splitext

logger = logging.getLogger(__name__)


def _get_http_response_size(resp: Response) -> Optional[int]:
    try:
        return int(resp.headers["content-length"])
    except (ValueError, KeyError, TypeError):
        return None


def _prepare_download(
    resp: Response,
    link: Link,
    progress_bar: str,
) -> Iterable[bytes]:
    total_length = _get_http_response_size(resp)

    if link.netloc == PyPI.file_storage_domain:
        url = link.show_url
    else:
        url = link.url_without_fragment

    logged_url = redact_auth_from_url(url)

    if total_length:
        logged_url = f"{logged_url} ({format_size(total_length)})"

    if is_from_cache(resp):
        logger.info("Using cached %s", logged_url)
    else:
        logger.info("Downloading %s", logged_url)

    if logger.getEffectiveLevel() > logging.INFO:
        show_progress = False
    elif is_from_cache(resp):
        show_progress = False
    elif not total_length:
        show_progress = True
    elif total_length > (512 * 1024):
        show_progress = True
    else:
        show_progress = False

    chunks = response_chunks(resp)

    if not show_progress:
        return chunks

    renderer = get_download_progress_renderer(bar_type=progress_bar, size=total_length)
    return renderer(chunks)


def sanitize_content_filename(filename: str) -> str:
    """
    Sanitize the "filename" value from a Content-Disposition header.
    """
    return os.path.basename(filename)


def parse_content_disposition(content_disposition: str, default_filename: str) -> str:
    """
    Parse the "filename" value from a Content-Disposition header, and
    return the default filename if the result is empty.
    """
    m = email.message.Message()
    m["content-type"] = content_disposition
    filename = m.get_param("filename")
    if filename:
        # We need to sanitize the filename to prevent directory traversal
        # in case the filename contains ".." path parts.
        filename = sanitize_content_filename(str(filename))
    return filename or default_filename


def _get_http_response_filename(resp: Response, link: Link) -> str:
    """Get an ideal filename from the given HTTP response, falling back to
    the link filename if not provided.
    """
    filename = link.filename  # fallback
    # Have a look at the Content-Disposition header for a better guess
    content_disposition = resp.headers.get("content-disposition")
    if content_disposition:
        filename = parse_content_disposition(content_disposition, filename)
    ext: Optional[str] = splitext(filename)[1]
    if not ext:
        ext = mimetypes.guess_extension(resp.headers.get("content-type", ""))
        if ext:
            filename += ext
    if not ext and link.url != resp.url:
        ext = os.path.splitext(resp.url)[1]
        if ext:
            filename += ext
    return filename


def _http_get_download(session: PipSession, link: Link) -> Response:
    target_url = link.url.split("#", 1)[0]
    resp = session.get(target_url, headers=HEADERS, stream=True)
    raise_for_status(resp)
    return resp


class Downloader:
    def __init__(
        self,
        session: PipSession,
        progress_bar: str,
    ) -> None:
        self._session = session
        self._progress_bar = progress_bar

    def __call__(self, link: Link, location: str) -> Tuple[str, str]:
        """Download the file given by link into location."""
        try:
            resp = _http_get_download(self._session, link)
        except NetworkConnectionError as e:
            assert e.response is not None
            logger.critical(
                "HTTP error %s while getting %s", e.response.status_code, link
            )
            raise

        filename = _get_http_response_filename(resp, link)
        filepath = os.path.join(location, filename)

        chunks = _prepare_download(resp, link, self._progress_bar)
        with open(filepath, "wb") as content_file:
            for chunk in chunks:
                content_file.write(chunk)
        content_type = resp.headers.get("Content-Type", "")
        return filepath, content_type


class BatchDownloader:
    def __init__(
        self,
        session: PipSession,
        progress_bar: str,
    ) -> None:
        self._session = session
        self._progress_bar = progress_bar

    def __call__(
        self, links: Iterable[Link], location: str
    ) -> Iterable[Tuple[Link, Tuple[str, str]]]:
        """Download the files given by links into location."""
        for link in links:
            try:
                resp = _http_get_download(self._session, link)
            except NetworkConnectionError as e:
                assert e.response is not None
                logger.critical(
                    "HTTP error %s while getting %s",
                    e.response.status_code,
                    link,
                )
                raise

            filename = _get_http_response_filename(resp, link)
            filepath = os.path.join(location, filename)

            chunks = _prepare_download(resp, link, self._progress_bar)
            with open(filepath, "wb") as content_file:
                for chunk in chunks:
                    content_file.write(chunk)
            content_type = resp.headers.get("Content-Type", "")
            yield link, (filepath, content_type)
